(** * Theory of Sets : Ordinals
  Copyright INRIA (2011-2013) Marelle Team (Jose Grimm).
*)

(* $Id: sset16.v,v 1.70 2017/03/14 16:48:55 grimm Exp $ *)

Require Import ssreflect ssrfun ssrbool eqtype ssrnat.
Require Export sset15 sset16a. 

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Module NumberSums.
  
Definition card_fifteen := \3c *c \5c.

Lemma nat_to_B15: nat_to_B 15 = card_fifteen.
Proof.
have pa: nat_to_B 3 = \3c  by  rewrite /=  succ_zero succ_one.
have pb:  nat_to_B 5 = \5c by rewrite /card_five /card_four - pa.
by rewrite -[15] /(3 * 5) nat_to_B_prod pa pb.
Qed.  

Definition ndsC n := csucc (n *c (\2c ^c (cpred n))).
Definition nds_fbd f:= forall n k, natp n ->  \2c <=c n ->
     \0c <c k -> k <c n ->
      (ndsC k) *c (f (n -c k)) <=c (f n).
Definition nds_fmax f:= forall n, natp n ->  \2c <=c n ->
   (exists k,  [/\ \0c <c k, k <c n & 
      (ndsC k) *c (f (n -c k)) = (f n)]).

Definition nds_fprops f:=
  [/\ f \0c = \1c, f \1c = \1c, nds_fbd f &  nds_fmax f].

Lemma NS_ndsC n: natp n -> natp (ndsC n).
Proof.
move => nN.
apply: NS_succ; apply: (NS_prod nN); apply: (NS_pow NS2 (NS_pred nN)).
Qed.

Lemma NS_nds f n: nds_fprops f -> natp n -> natp (f n).
Proof.
move => [h0 h1 _ hm] nN.
move: (cleR (CS_nat nN));move:n nN {1 3} (n); apply: Nat_induction.
  move => n /cle0 ->; rewrite h0; fprops.
move => m mN Hrec n / cle_eqVlt;case; last by  move/(cltSleP mN); apply: Hrec.
move ->.  
case: (equal_or_not m \0c) => mp; first by  rewrite mp succ_zero h1; fprops.
have smN := (NS_succ mN).
have sm2: \2c <=c csucc m.
  apply: (cge2 (CS_succ m) (@succ_nz m)); rewrite - succ_zero.
  by move/ (succ_injective1 (CS_nat mN) CS0).
move: (hm (csucc m) smN sm2) => [k [kp ksm] <-].
move:(NS_lt_nat ksm smN) => kN.
have  aux: (csucc m -c k) <=c m.
  apply:(csum_le2l kN (NS_diff _ smN) mN); rewrite (cdiff_pr (proj1 ksm)).
  apply/(cleSltP mN); rewrite csumC; apply:(csum_M0lt mN (nesym (proj2 kp))).
exact: NS_prod (NS_ndsC kN) (Hrec _ aux).
Qed.

Definition nds_explicit C4 C5 C6 k i :=
  Yo (i = \0c) (C5 ^c k)
    (Yo (i = \1c) (C6 *c C5 ^c (k -c \1c)) 
       (Yo (i = \2c) (C6 ^c \2c *c C5 ^c (k -c \2c))
          (Yo (i = \3c)
              (Yo (k = \1c) (C4 ^c \2c)
                  (Yo (k = \2c) (C4 ^c \2c  *c C5)
                      (C6 ^c \3c *c C5 ^c (k -c \3c))))
              (C4 *c C5 ^c k)))).

(* old version to be restored 

Lemma  nds_thm_bourbaki f: nds_fprops f ->
  [/\ (forall k, natp k -> card_fifteen <=c k -> 
      f(k +c \5c) = (ndsC \5c) *c f k), 
   (forall n, natp n -> \3c <=c n ->
       exists k, [/\ \2c <=c k, k <=c \7c, k <c n &
         f n = ndsC k *c f (n -c k)]),
   (forall n, natp n -> n <=c \7c -> f (csucc n) = ndsC n) &
   forall n, natp n -> f (csucc n) = Yo (n <=c \7c) (ndsC n)
     (nds_explicit (ndsC \4c) (ndsC \5c)(ndsC \6c) (n %/c \5c) (n %%c \5c)) ]. 
Proof.
move => h0; move: (h0) => [_ ha hb hc].
pose c := Nds.ndsC.
have cv n: nat_to_B (c n) = ndsC (nat_to_B n).
  rewrite /c/Nds.ndsC/ndsC; case:n; first by rewrite cprod0l.
  move => n /=. rewrite (cpred_pr2 (nat_to_B_Nat n)).
  by rewrite nat_to_B_prod nat_to_B_pow /= succ_zero succ_one.
pose F n := B_to_nat (f (nat_to_B n)).
have fp1 n: nat_to_B (F n) =  f (nat_to_B n).
  by rewrite /F (B_to_nat_pa (NS_nds h0 (nat_to_B_Nat n))).
have NB0: nat_to_B 0 = \0c by [].
have NB1: nat_to_B 1 = \1c by rewrite /= succ_zero.
have NB2: nat_to_B 2 = \2c by rewrite /= succ_zero succ_one .
have NB3: nat_to_B 3 = \3c by rewrite /= succ_zero succ_one .
have NB4: nat_to_B 4 = \4c by rewrite /= succ_zero succ_one .
have NB5: nat_to_B 5 = \5c by  rewrite /card_five - NB4.
have NB6: nat_to_B 6 = \6c by  rewrite /card_six - NB5. 
have NB7: nat_to_B 7 = \7c by  rewrite /card_seven - NB6. 
move: (fp1 1); rewrite NB1 ha - NB1 => /nat_to_B_injective fp2.
have fp3: forall n k, 1 < n -> 0 < k < n -> c k * F (n - k) <= F n.
  move => n k l1n /andP [l0k l2n].
  move/nat_to_B_le: l1n;  move/nat_to_B_lt: l0k; move/nat_to_B_lt: l2n.
  rewrite NB2 NB0 => lkn l0k l2n.
  move: (hb _ _ (nat_to_B_Nat n) l2n l0k lkn).
  rewrite  - nat_to_B_diff - fp1 - fp1 - cv - nat_to_B_prod.
  by move/nat_to_B_le.
have fp4 n:  1 < n -> exists2 k : nat, 0 < k < n & F n = c k * F (n - k).
  have nN :=  (nat_to_B_Nat n).
  move/nat_to_B_le; rewrite NB2 => n2.
  move: (hc _ nN  n2) => [K [ka kb kc]].
  have kN:= (NS_lt_nat kb nN).
  move: (B_to_nat_pa kN); set k := B_to_nat _ => kv.
  exists k; first by  apply/andP; split; apply/nat_to_B_lt; rewrite kv //.
  apply: nat_to_B_injective.
  by rewrite fp1 nat_to_B_prod cv fp1 nat_to_B_diff kv kc.
move:(Nds.nds_thm fp2 fp3 fp4) => [ra rd rb rc].
have r4: forall n, natp n -> n <=c \7c -> f (csucc n) = ndsC n.
  move => n nN.
  move: (B_to_nat_pa nN) => nv.
  rewrite - NB7 - nv => /nat_to_B_le /ra.
  by rewrite nat_to_B_succ - fp1 - cv -/c => <-.
split.
-  move => k kN.
   move: (B_to_nat_pa kN) => kv.
   rewrite - nat_to_B15 - kv => /nat_to_B_le /rc.
   by rewrite - NB5 - nat_to_B_sum - fp1 - fp1 - cv - nat_to_B_prod => <-.
- move => n nN.
  move: (B_to_nat_pa nN) => nv.
  rewrite - NB3 - NB7 - nv => /nat_to_B_le /rd.
  move => [k [/andP[ka kb] kc kd]]; exists (nat_to_B k); split => //.
  + by rewrite -NB2; apply/nat_to_B_le.
  + by apply/nat_to_B_le.
  + by apply/nat_to_B_lt.
  + by rewrite -fp1 kd - nat_to_B_diff - fp1 - cv nat_to_B_prod.
- done.
- move => n nN.
  move: (B_to_nat_pa nN) => nv.
  rewrite - nv nat_to_B_succ - fp1 rb.
  case: (ltnP 7 (B_to_nat n)) => lt1; last first.
    move/nat_to_B_le: lt1; rewrite NB7 => lt1.
    by rewrite (Y_true lt1) cv.
  move/nat_to_B_lt: lt1; rewrite NB7 => /cltNge lt1; rewrite (Y_false lt1).
  rewrite - NB5; move: (nat_to_B_quorem (B_to_nat n) 5) => [<- <-].

  have NBI a b : a <> b -> nat_to_B a <> nat_to_B b. 
    by move => h /nat_to_B_injective.
  rewrite /nds_explicit.
  rewrite - NB0;case: eqP => ci0.
    by rewrite ci0 (Y_true (erefl _)) nat_to_B_pow cv. 
  rewrite (Y_false (NBI _ _ ci0)) - NB1;case: eqP => ci1.
    rewrite ci1 (Y_true (erefl _)) nat_to_B_prod cv NB6 - subn1.
    by rewrite nat_to_B_pow nat_to_B_diff  cv. 
  rewrite (Y_false (NBI _ _ ci1)) - NB2;case: eqP => ci2.
    rewrite ci2 (Y_true (erefl _)) nat_to_B_prod 2! nat_to_B_pow.
    by rewrite - subn2 cv cv NB5 NB6 nat_to_B_diff. 
  rewrite (Y_false (NBI _ _ ci2)) - NB3;case: eqP => ci3; last first.
    rewrite (Y_false (NBI _ _ ci3)) nat_to_B_prod nat_to_B_pow. 
    by rewrite cv cv NB4 NB5.
  rewrite ci3  (Y_true (erefl _)); case: eqP => ck1.
    by rewrite ck1 (Y_true (erefl _)) nat_to_B_pow cv NB4.
  rewrite (Y_false (NBI _ _ ck1)) fun_if. 
  rewrite 2! nat_to_B_prod  3! nat_to_B_pow cv cv cv NB2 NB3 NB4 NB5 NB6.
  case: eqP => ck2.
    by rewrite ck2 NB2 (Y_true (erefl _)).
  rewrite - {1} NB2 (Y_false (NBI _ _ ck2)).
  by rewrite nat_to_B_diff NB3.
Qed.

 *)

Lemma  nds_thm_bourbaki f: nds_fprops f ->
  [/\ (forall k, natp k -> card_fifteen <=c k -> 
      f(k +c \5c) = (ndsC \5c) *c f k), 
   (forall n, natp n -> n <=c \7c -> f (csucc n) = ndsC n) &
   forall n, natp n -> f (csucc n) = Yo (n <=c \7c) (ndsC n)
     (nds_explicit (ndsC \4c) (ndsC \5c)(ndsC \6c) (n %/c \5c) (n %%c \5c)) ]. 
Proof.
move => h0; move: (h0) => [_ ha hb hc].
pose c := Nds.ndsC.
have cv n: nat_to_B (c n) = ndsC (nat_to_B n).
  rewrite /c/Nds.ndsC/ndsC; case:n; first by rewrite cprod0l.
  move => n /=. rewrite (cpred_pr2 (nat_to_B_Nat n)).
  by rewrite nat_to_B_prod nat_to_B_pow /= succ_zero succ_one.
pose F n := B_to_nat (f (nat_to_B n)).
have fp1 n: nat_to_B (F n) =  f (nat_to_B n).
  by rewrite /F (B_to_nat_pa (NS_nds h0 (nat_to_B_Nat n))).
have NB0: nat_to_B 0 = \0c by [].
have NB1: nat_to_B 1 = \1c by rewrite /= succ_zero.
have NB2: nat_to_B 2 = \2c by rewrite /= succ_zero succ_one .
have NB3: nat_to_B 3 = \3c by rewrite /= succ_zero succ_one .
have NB4: nat_to_B 4 = \4c by rewrite /= succ_zero succ_one .
have NB5: nat_to_B 5 = \5c by  rewrite /card_five - NB4.
have NB6: nat_to_B 6 = \6c by  rewrite /card_six - NB5. 
have NB7: nat_to_B 7 = \7c by  rewrite /card_seven - NB6. 
move: (fp1 1); rewrite NB1 ha - NB1 => /nat_to_B_injective fp2.
have fp3: forall n k, 1 < n -> 0 < k < n -> c k * F (n - k) <= F n.
  move => n k l1n /andP [l0k l2n].
  move/nat_to_B_le: l1n;  move/nat_to_B_lt: l0k; move/nat_to_B_lt: l2n.
  rewrite NB2 NB0 => lkn l0k l2n.
  move: (hb _ _ (nat_to_B_Nat n) l2n l0k lkn).
  rewrite  - nat_to_B_diff - fp1 - fp1 - cv - nat_to_B_prod.
  by move/nat_to_B_le.
have fp4 n:  1 < n -> exists2 k : nat, 0 < k < n & F n = c k * F (n - k).
  have nN :=  (nat_to_B_Nat n).
  move/nat_to_B_le; rewrite NB2 => n2.
  move: (hc _ nN  n2) => [K [ka kb kc]].
  have kN:= (NS_lt_nat kb nN).
  move: (B_to_nat_pa kN); set k := B_to_nat _ => kv.
  exists k; first by  apply/andP; split; apply/nat_to_B_lt; rewrite kv //.
  apply: nat_to_B_injective.
  by rewrite fp1 nat_to_B_prod cv fp1 nat_to_B_diff kv kc.
move:(Nds.nds_thm fp2 fp3 fp4) => [ra rd rb rc].
have r4: forall n, natp n -> n <=c \7c -> f (csucc n) = ndsC n.
  move => n nN.
  move: (B_to_nat_pa nN) => nv.
  rewrite - NB7 - nv => /nat_to_B_le /ra.
  by rewrite nat_to_B_succ - fp1 - cv -/c => <-.
split.
-  move => k kN.
   move: (B_to_nat_pa kN) => kv.
   rewrite - nat_to_B15 - kv => /nat_to_B_le /rc.
   by rewrite - NB5 - nat_to_B_sum - fp1 - fp1 - cv - nat_to_B_prod => <-.
- done.
- move => n nN.
  move: (B_to_nat_pa nN) => nv.
  rewrite - nv nat_to_B_succ - fp1 rb. rewrite/Nds.nds_value.
  case: (ltnP 7 (B_to_nat n)) => lt1; last first.
    move/nat_to_B_le: lt1; rewrite NB7 => lt1.
    by rewrite (Y_true lt1) cv.  
  move/nat_to_B_lt: lt1; rewrite NB7 => /cltNge lt1; rewrite (Y_false lt1).
  rewrite - NB5; move: (nat_to_B_quorem (B_to_nat n) 5) => [<- <-].

  have NBI a b : a <> b -> nat_to_B a <> nat_to_B b. 
    by move => h /nat_to_B_injective.
  rewrite /nds_explicit.
  rewrite - NB0;case: eqP => ci0.
    by rewrite ci0 (Y_true (erefl _)) nat_to_B_pow cv. 
  rewrite (Y_false (NBI _ _ ci0)) - NB1;case: eqP => ci1.
    rewrite ci1 (Y_true (erefl _)) nat_to_B_prod cv NB6 - subn1.
    by rewrite nat_to_B_pow nat_to_B_diff  cv. 
  rewrite (Y_false (NBI _ _ ci1)) - NB2;case: eqP => ci2.
    rewrite ci2 (Y_true (erefl _)) nat_to_B_prod 2! nat_to_B_pow.
    by rewrite - subn2 cv cv NB5 NB6 nat_to_B_diff. 
  rewrite (Y_false (NBI _ _ ci2)) - NB3;case: eqP => ci3; last first.
    rewrite (Y_false (NBI _ _ ci3)) nat_to_B_prod nat_to_B_pow. 
    by rewrite cv cv NB4 NB5.
  rewrite ci3  (Y_true (erefl _)); case: eqP => ck1.
    by rewrite ck1 (Y_true (erefl _)) nat_to_B_pow cv NB4.
  rewrite (Y_false (NBI _ _ ck1)) fun_if. 
  rewrite 2! nat_to_B_prod  3! nat_to_B_pow cv cv cv NB2 NB3 NB4 NB5 NB6.
  case: eqP => ck2.
    by rewrite ck2 NB2 (Y_true (erefl _)).
  rewrite - {1} NB2 (Y_false (NBI _ _ ck2)).
  by rewrite nat_to_B_diff NB3.
Qed.


(** Number of decompositions *)

Definition nds_sc n (X:fterm) g := osumf (fun z => (X (Vf g z))) n. 

Definition nds_sums n (X:fterm) := fun_image (permutations n)  (nds_sc n X).

Definition nds_card n (X: fterm) := cardinal (nds_sums n X).

Definition nds_ax := ord_below_n.

Definition small_ordinal x := x <o oopow omega0.

Lemma small_ordinaP x: ordinalp x -> (small_ordinal x <-> odegree x <o omega0).
Proof.
move => ox.
case: (ozero_dichot ox) => xp.
  rewrite xp - oopow1/small_ordinal odegree_zero.
  split => _; apply: oopow_pos; fprops.
move: (the_CNF_p4 xp) => [le1 lt1]; split => hyp.
  by move/(opow_Mo_ltP (OS_degree ox) OS_omega): (ole_ltT le1 hyp). 
apply: (olt_ltT lt1); apply:opow_Mo_lt.
move/limit_ordinal_P:omega_limit => [_ h2]; exact: (h2 _  hyp).
Qed.


Definition rel_small_ordinal d c x:=
  odegree x <=o d /\ let X := the_CNF x in
  forall i, i <c (P X) -> Vg (Q (Q X)) i <=o c.

Lemma rel_small_prop1 d c x: d <o omega0 -> ordinalp x ->
  rel_small_ordinal d c x -> small_ordinal x.
Proof.
by move => ha ox [hb _]; apply/(small_ordinaP ox); apply: (ole_ltT hb ha).
Qed.

Lemma rel_small_zero d c: ordinalp d -> ordinalp c -> rel_small_ordinal d c \0o.
Proof.
move => od oc; split.
  by rewrite odegree_zero; apply: ole0x.
by move => /= i ; rewrite the_CNF_p1 =>  /clt0.
Qed.

Lemma rel_small_prop2 md mc e c n (x := (CNFbvo e c n)):
  CNFb_axo e c n -> natp n -> 
  (forall i, i <c n -> c i <=o mc) ->
  e (cpred n) <=o md ->
  rel_small_ordinal md mc x.
Proof.
move => ax nN s1 s2.
have cx: the_CNF x = CNFbB n e c by exact:(the_CNF_p3 ax nN).
rewrite /rel_small_ordinal /odegree cx.
rewrite /CNFbB; aw; split.
  move: (NS_pred nN) => pN; bw; Ytac h => //;  apply: (ole0x (proj32 s2)).
by move => i lin; move: (NS_lt_nat lin nN) => iN; bw; Ytac0; apply: s1.
Qed.

Lemma rel_small_oopowc e c: ordinalp e -> \0o <o c -> c <o omega0 ->
  rel_small_ordinal e c ((oopow e) *o c).
Proof.
move => oe cp cl.
set ee := fun i:Set => e; set cc := fun i:Set => c.
have ax: CNFb_axo ee cc \1c.
  split => //; split => //; first exact ole_2omega.
  by move => i _ /= /clt1 /succ_nz.
have bd: (forall i, i <c \1c -> cc i <=o c).
  rewrite /cc;move => i _ /=; apply: (oleR (proj31_1 cl)).
have be: ee (cpred \1c) <=o e by  apply:(oleR oe).
move: (rel_small_prop2 ax NS1 bd be). 
rewrite /CNFbvo /CNFbv osum_f1 //.
exact: (OS_prod2 (OS_oopow oe) (proj31_1 cl)). 
Qed.


Lemma rel_small_oopow n: ordinalp n -> rel_small_ordinal n \1o (oopow n).
Proof.
move => on; move:(rel_small_oopowc on olt_01 olt_1omega).
by rewrite (oprod1r (OS_oopow on)).
Qed.

Lemma rel_small_one:rel_small_ordinal \0c \1o \1o.
Proof. by move: (rel_small_oopow OS0); rewrite oopow0. Qed.

Lemma rel_small_omega: rel_small_ordinal \1c \1o omega0.
Proof. by move: (rel_small_oopow OS1); rewrite oopow1. Qed.


Definition nds_small_ax (X:fterm) n := forall k , k <c n -> small_ordinal (X k).
Definition nds_rel_small_ax d c (X:fterm) n:=
  forall k , k <c n -> rel_small_ordinal d c  (X k).
Definition nds_same_dc_bound n (X Y: fterm) :=
  forall d c, ordinalp d -> ordinalp c ->
              (nds_rel_small_ax d c X n  <-> nds_rel_small_ax d c Y n).

Lemma rel_small_le1 d1 d2 c x:
  d1 <=o d2 -> rel_small_ordinal d1 c x -> rel_small_ordinal d2 c x.
Proof.
move => le1 [ha hb]; split => //; apply:(oleT ha le1).
Qed.

Lemma rel_small_le2 d c1 c2 x:
  c1 <=o c2 -> rel_small_ordinal d c1 x -> rel_small_ordinal d c2 x.
Proof.
by move => le1 [ha hb]; split => //= i /hb le2; apply:(oleT le2 le1).
Qed.

Lemma nds_same_dc_bound_perm n X f (Y:= fun i =>X (Vf f i)):
  natp n -> inc f (permutations n) ->  nds_same_dc_bound n X Y.
Proof.
move => nN /permutationsP [[[ff _] [_ sjf]] sf tf] => d c od oc.
rewrite/Y;split => Hx i lin.
  by apply: Hx; apply/(NltP nN); Wtac; rewrite sf;apply/(NltP nN). 
have /sjf [j js ->] : inc i (target f) by rewrite tf;apply/(NltP nN).
by apply:Hx;apply/(NltP nN); rewrite - sf.
Qed.

Lemma nds_card_bd n X: natp n ->  nds_card n X <=c (factorial n).
Proof.
move=> nN.
rewrite -{2} (card_nat nN) -(number_of_permutations (finite_set_nat nN)).
apply: (fun_image_smaller). 
Qed.

Lemma nds_ax_perm X n f: natp n -> nds_ax X n -> 
   inc f (permutations n) ->
   (nds_ax (fun z => X (Vf f z)) n  /\  ordinalp (nds_sc n X f)).
Proof.
move=> nN ax => /Zo_P [] /functionsP [ff sf tf] bf.
have pc: nds_ax (fun z => X (Vf f z)) n.
  move => t /(NltP nN) tn; apply: ax; apply/(NltP nN); Wtac. 
by split => //; apply: OS_osumf.
Qed.

Lemma nds_sums_oset X n: natp n -> nds_ax X n -> ordinal_set (nds_sums n X).
Proof.
move => nN ax x /funI_P [z zp ->].
exact: proj2 (nds_ax_perm  nN ax zp).
Qed.
  
Lemma nds_sc_exten X X' n f: 
  natp n -> (same_below X X' n) -> 
  inc f (permutations n) ->
  nds_sc n X f = nds_sc n X' f.
Proof.
move =>nN h /Zo_P [] /functionsP [fg sg tg] bg;apply: (osumf_exten nN).
move => i /(NltP nN) ia; apply: h; apply/(NltP nN); Wtac.
Qed.


Lemma nds_sums_exten m Y1 Y2: natp m -> same_below Y1 Y2 m ->
   nds_sums m Y1 = nds_sums m Y2. 
Proof.
move => ha hb;set_extens t;
   move => /funI_P [z zp ->];apply/funI_P; ex_tac. 
by apply:nds_sc_exten.
by symmetry;apply:nds_sc_exten.
Qed.

Lemma nds_card_exten m Y1 Y2:  natp m -> same_below Y1 Y2 m ->
   nds_card m Y1 = nds_card m Y2. 
Proof.
by move => ha hb; rewrite /nds_card (nds_sums_exten ha hb).
Qed.

Lemma nds_card_perm_inv n X g:
  natp n -> nds_ax X n -> inc g (permutations n) -> 
  nds_card n X = nds_card n (fun z => X (Vf g z)).
Proof.
move => nN ax /Zo_P [] /functionsP [fg sg tg] bg.
rewrite /nds_card /nds_sums; apply: f_equal.
set E := permutations n.
set_extens t; move /funI_P => [z ze ->]; apply /funI_P.
  move: (inverse_bij_fb bg) => ibg.
  move: ze => /Zo_P [] /functionsP [fz sz tz] bz.
  have fa: (inverse_fun g) \coP z by  red;aw;split => //; [fct_tac | ue].
  have fb: function  ((inverse_fun g) \co z) by fct_tac.
  exists ((inverse_fun g) \co z).
  apply: Zo_i; [ apply /functionsP;red;aw;split => // | by apply compose_fb ].
  apply:(osumf_exten nN) => i lin.
  have iz: inc i (source z)  by rewrite sz; apply /(NltP nN i).
  have cp: inverse_fun g \coP z by split; [ fct_tac | exact | aw; ue].
  have fzy: inc (Vf z i) (target g) by rewrite tg - tz; Wtac.
  by rewrite (compf_V cp iz) (inverse_V bg fzy).
move: ze =>  /Zo_P [] /functionsP [fz sz tz] bz.
have fa: g \coP z by red;aw;split => //; ue.
have fb: function  (g \co z) by fct_tac.
exists (g \co z).
  apply: Zo_i; [ apply /functionsP;red;aw;split => // | by apply compose_fb ].
apply:(osumf_exten nN) => i lin.
by rewrite compf_V // sz; apply/(NltP nN i).
Qed.

Lemma NS_nds_card n X: natp n -> natp (nds_card n X).
Proof.
move=> nN; move: (nds_card_bd X nN) => pa.
exact:(NS_le_nat pa (NS_factorial nN)).
Qed.

Lemma nds_card_nz n X:  \1c <=c nds_card n X.
Proof.
rewrite /nds_card/nds_sums; apply: cge1; first by fprops. 
apply: card_nonempty1; apply:funI_setne.
exists (identity n).
apply: Zo_i; [ apply /functionsP;red;aw;split;fprops | apply identity_fb].
Qed.

Lemma nds_card_01 X n: (n = \0c \/ n = \1c) ->  nds_card n X = \1c.
Proof.
move=> np.
have nN: natp n by case: np => ->; fprops.
apply: cleA (nds_card_nz n X).
move: (nds_card_bd X nN).
by case: np => ->; rewrite ? factorial0 ? factorial1.
Qed.

Lemma nds_card2_aux X: nds_ax X \2c -> osumf X \2c = (X \1c) +o (X \0c).
Proof.
rewrite - succ_one => pa.
rewrite (osum_fS _ NS1) (osum_f1) //;apply: pa.
rewrite succ_one; apply: clt_02.
Qed.

Lemma nds_sc_id n X (f := identity n) :
  natp n ->  nds_ax X n ->
  (inc f (permutations n) /\ (nds_sc n X f = osumf X n)).
Proof.
move=> nN ax.
have bi: bijection f by apply identity_fb.
split. 
  by rewrite /f; apply: Zo_i => //; apply /functionsP; red;aw;split;fprops.
apply:(osumf_exten nN) => i lin.
by rewrite /f identity_V //; apply/(NltP nN).
Qed.

Lemma nds_card2 (X:=(variant \0c \1o omega0)):
    nds_ax X \2c /\ nds_card \2c X = \2c.
Proof.
move: osum2_nc => /= nc.
have nX: nds_ax X \2c by move => t t2; rewrite /X /variant; Ytac h; fprops.
split => //; apply: cleA.
  by move: (nds_card_bd X NS2); rewrite factorial2.
move: (nds_sc_id NS2 nX) => [ha hb].
have ne:= C1_ne_C0.
apply /cle2P. 
exists (\1o +o omega0), (omega0 +o \1o); split => //;apply /funI_P; last first.
  exists (identity \2c) => //.
  by rewrite hb (nds_card2_aux nX) /X /variant; Ytac0; Ytac0.
rewrite - (proj2 (proj2  constants_v)).
move: inc_C0_C2 inc_C1_C2 C2_P => i9 i1 iv.
have ta: lf_axiom (variant C0 C1 C0) C2 C2.
  by move => t _; rewrite /variant; Ytac h.
have fp: inc (Lf (variant C0 C1 C0) C2 C2) (permutations C2).
  apply/Zo_i; first by apply /functionsP; split; [apply: lf_function | aw | aw].
  apply: lf_bijective => //.
    move=> u v us vs. 
    rewrite /variant; case /iv: us; case /iv: vs;
       move => -> ->; Ytac0 => //; Ytac0 => //.
    move => y yv; case/iv: yv; move => ->.
      by exists C1 => //; rewrite /variant; Ytac0.
      by exists C0 => //; rewrite /variant; Ytac0.
have ax2:= (proj1 (nds_ax_perm NS2 nX fp)).
exists (Lf (variant C0 C1 C0) C2 C2); first exact.
rewrite /nds_sc nds_card2_aux; aw; aw; rewrite /X /variant; Ytac0.
rewrite Y_true // Y_false // Y_true //.
Qed.

Lemma nds_card3 (A:= omega0 ^o \2o) (B:= omega0) (C := omega0 +o \1o) 
  (s1 := A +o (B +o C)) (s2:= A +o (C +o B)) (s3 := B +o (A +o C))
  (s4 := B +o (C +o A)) (s5:= C +o (A +o B)) (s6 := C +o (B +o A)):
  cardinal (((doubleton s1 s2) \cup (doubleton s3 s5)) \cup
    (doubleton s4 s6)) = \5c.
Proof.
set T := union2 _ _.
move: OS_omega OS1 OS2 => oB o1 o2.
have oA: ordinalp A by apply: OS_pow.
have oC: ordinalp C by apply: OS_sum2.
move: (ord_negl_p4 olt_02); rewrite oopow0 -/A => pa.
move: (ord_negl_p4 olt_12); rewrite oopow1 -/A -/B => pb.
have e1: s3 = A +o C by rewrite /s3 (osumA oB oA oC) pb.
have e2: s6 = C +o A by rewrite /s6  pb.
have pc: C +o A = A by rewrite - (osumA oB OS1 oA)  pa pb.
have pd: (doubleton s4 s6) = singleton A by rewrite /s4 e2 pc pb.
have e3: s2 = A +o (B +o B). 
   by rewrite /s2 - (osumA oB  o1 oB)  (osum_int_omega olt_1omega).
have e4: s5= A +o B by rewrite /s5  (osumA oC oA oB) pc. 
set T1:=  (doubleton s1 s2) \cup (doubleton s3 s5).
have bz: B <> \0o by apply: omega_nz.
have onz: \1o <> \0o by fprops.
case: (equal_or_not C \0o) => cz; first by move: (osum2_zero oB OS1 cz) => [_].
have nAt: ~ (inc A T1).
  have sA: forall b, ordinalp b -> A = A +o b -> b = \0o. 
     move => b ob h; symmetry in h; exact (osum2_a_ab oA ob h).
  case /setU2_P; case /set2_P.
  - by move/(sA _ (OS_sum2 oB oC)) => /(osum2_zero oB oC) [_].
  - by move/(sA _ (OS_sum2 oC oB)) => /(osum2_zero oC oB) [].
  - by rewrite e1; move /(sA _ oC).
  - by rewrite e4; move /(sA _ oB).
have ->: T = T1 +s1 A by rewrite -pd //.
rewrite /card_five (csucc_pr nAt); congr csucc.
set T2 := (doubleton s1 s2) +s1 s3.
have ->: T1 = T2 +s1 s5.
  set_extens t.
     case /setU2_P => h; first by apply /setU1_P; left; apply /setU1_P; left.
     case /set2_P :h => h; first by apply /setU1_P; left; apply /setU1_P; right.
     by apply /setU1_P; right.
  case /setU1_P =>h; last by  apply /setU2_P;  right; apply/set2_P; right.
  case /setU1_P:h => h; last by  apply /setU2_P; right; apply/set2_P; left.
  by apply /setU2_P; left.
have aux: forall b c, ordinalp b -> ordinalp c -> A +o b = A +o c -> b = c.
  move => b c ob oc ; apply: (osum2_simpl ob oc oA).
have auB: forall b c, ordinalp b -> ordinalp c -> B +o b = B +o c -> b = c.
  move => b c ob oc ; apply: (osum2_simpl ob oc oB).
case (equal_or_not  B C) => nCB; first by move: (osum2_a_ab oB o1 (esym nCB)).
have nat2: ~ inc s5 T2.
  rewrite /T2 e4 /s1 e1 e3 => /setU1_P; case => h; last first.
    case: nCB; exact (aux _ _ oB oC h).
  case /set2_P:h=> h. 
       move: (aux _ _ oB (OS_sum2 oB oC) h) => h1; symmetry in h1.
       case: cz; exact: (osum2_a_ab oB oC h1).
     move: (aux _ _ oB (OS_sum2 oB oB) h) => h1; symmetry in h1.
     case: bz; exact: (osum2_a_ab oB oB h1).
move: (olt_1omega) => [h2 h3].
move: (oltS oB) => h4.
have nat3: ~ inc s3 (doubleton s1 s2).
  rewrite /s1 e1 e3;case  /set2_P=> h.
    move: (aux _ _ oC (OS_sum2 oB oC) h) => h1.
    move: (ole_ltT h2 h4) => [_ h5].
    by move: (auB _ _ o1 oC h1); rewrite /C (osucc_pr oB).
  move: (aux _ _ oC (OS_sum2 oB oB) h) => h1.
  case: h3; exact: (auB _ _ o1 oB h1).
rewrite /card_four (csucc_pr nat2); apply: f_equal.
rewrite /card_three (csucc_pr nat3); apply: f_equal.
apply: cardinal_set2.
rewrite /s1 e3 => h; move: (aux _ _ (OS_sum2 oB oC) (OS_sum2 oB oB) h) => h1.
by case: nCB; rewrite (auB _ _ oC oB h1).
Qed.

Definition nds_F n := 
 \osup (Zo Nat (fun z => exists2 X, nds_ax X n & nds_card n X = z)).

Definition nfs_F_small n :=
  exists X, [/\  nds_small_ax X n,  nds_ax X n & nds_card n X = nds_F n].
Definition nfs_F_rel_small d c n :=
  exists X, [/\ nds_rel_small_ax d c X n,  nds_ax X n & nds_card n X = nds_F n].

Lemma Nat_sup_pr T (s:= \osup T) k: 
  sub T Nat -> natp k -> (forall i, inc i T -> i <=c k) ->
  [/\ natp s, s <=c k,
    (forall i, inc i T -> i <=c s) &
    (T = emptyset \/ inc s T)].
Proof.
move=> TN kN km.
have ok:= (CS_nat kN).
have pa: cardinal_set T by move=> t tT; apply: CS_nat; apply: TN.
have pa': ordinal_set T by move=> t tT; apply: OS_nat; apply: TN.
have pb: s <=c k.
  apply: card_ub_sup => // i iT; move: (km _ iT);  apply: ocle.
have pc: natp s by apply: (NS_le_nat pb kN).
split => //.
  by move=> i iT; apply: card_sup_ub.
case: (emptyset_dichot T) => tne; [ by left | right].
case: (ord_sup_inVlimit pa' tne) => // li.
by move/olt_omegaP: pc => /(oleNgt (limit_ge_omega li)).
Qed.

Lemma nds_f_def n (f := nds_F n): natp n ->
  [/\ natp f,
   f <=c factorial n,
   (exists2 X, nds_ax X n & nds_card n X = f) &
   (forall X, nds_ax X n -> nds_card n X <=c f)].
Proof.
move => nN.
set T := (Zo Nat(fun z => exists2 X, nds_ax X n & nds_card n X = z)).
have ta:  sub T Nat by apply: Zo_S.
move: (NS_factorial nN) => tb.
have tc: (forall i, inc i T -> i <=c (factorial n)).
  rewrite /T;move=> i /Zo_P [iN [X ea <-]]; exact: (nds_card_bd X nN). 
move: (Nat_sup_pr ta tb tc) => [pa pb pc pd].
have pe X: nds_ax X n -> inc (nds_card n X) T.
  by move => ax; apply: Zo_i;[ apply: NS_nds_card | exists X].
case:pd => h.
  set X := (fun z: Set => \0o). 
  have aX: nds_ax X n by rewrite /X;move => i _ /=; fprops.
  by empty_tac1 (nds_card n X); apply: pe.
by split => //; [ move:h => /Zo_P [_] | move=> X xv; apply: pc;  apply: pe].
Qed.

Lemma nds_f01: (nds_F \0c = \1c /\ nds_F \1c = \1c).
Proof.
move: nds_card_01 => h.
move: (nds_f_def NS0); set f := nds_F \0c;move => [_ _ [X _ <-] _].
move: (nds_f_def NS1); set g := nds_F \0c;move => [_ _ [Y _ <-] _].
have pa: \0c = \0c \/ \0c = \1c by left.
have pb: \1c = \0c \/ \1c = \1c by right.
by rewrite (h X _ pa) (h Y _ pb).
Qed.

Lemma nds_f2: nds_F \2c = \2c.
Proof.
move: (nds_f_def NS2); set f := nds_F \2c.
move => [p1 p2 p3 p4].
move: nds_card2 =>  [ax xv]; move: (p4 _ ax); rewrite xv;  apply:cleA. 
by rewrite factorial2 in p2.
Qed.

Lemma nds_f_small_prop d c n: d <o omega0 ->
  nfs_F_rel_small d c n  -> nfs_F_small n. 
Proof.
move => dN [X [xA Xb Ac]]; exists X; split => // i lin.
by move: (xA _ lin); apply: (rel_small_prop1 dN (Xb _ lin)).
Qed.

Lemma nds_f0_small: nfs_F_rel_small \0o \1o \0c. 
Proof.
move: nds_card_01 => h.
move: (nds_f_def NS0); set f := nds_F \0c;move => [_ _ [X _ <-] _].
have pa: \0c = \0c \/ \0c = \1c by left.
exists (fun i => \0o); split; try  by move => i /clt0.
by rewrite (h _  _ pa) (proj1 nds_f01).
Qed.

Lemma nds_f1_small: nfs_F_rel_small \0o \1o \1c. 
Proof.
move: nds_card_01 => h.
move: (nds_f_def NS1); set f := nds_F \0c;move => [_ _ [X _ <-] _].
have pa: \1c = \0c \/ \1c = \1c by  right.
exists (fun i => \0o); split.
- move => i _; apply:(rel_small_zero OS0 OS1).
- move => i _; apply: OS0.
- by rewrite (h _  _ pa) (proj2 nds_f01).
Qed.


Lemma nds_f2_small: nfs_F_rel_small \1o \1o \2c. 
Proof.
move:  nds_card2. set X := variant _ _ _; move => [ha hb]; exists X; split.
- rewrite /X; move => i /clt2; case; move ->.
     rewrite variant_true1 - {3} oopow0; apply:(rel_small_le1 ole_01).
     exact:(rel_small_oopow OS0).
  rewrite variant_false1 - oopow1; apply:(rel_small_oopow OS1).
- exact.
- by rewrite hb nds_f2.
Qed.


Lemma osum_of_nat n X: natp n -> (forall i, i <c n -> natp (X i)) ->
  osumf X n = csumb n X.
Proof.
move: n; apply: Nat_induction.
  move =>  _; rewrite osum_f0 /csumb csum_trivial //; bw.
move => m mN Hrec zp; rewrite (osum_fS _ mN)  (succ_of_nat mN).
have ha: forall i, i <c m -> natp (X i). 
  move => i lim; apply: zp; exact: (clt_ltT lim (cltS mN)).
  have zmN:= (zp _ (cltS mN)); move: (nat_irreflexive mN) => mm.
  have sN: natp (csumb m X). 
    apply:finite_sum_finite; split.
      by hnf; bw;move => i lim; bw; apply: ha; apply /(NltP mN).
    by bw; apply: finite_set_nat.
  by rewrite (csumA_setU1 X mm) (Hrec ha) csumC (osum2_2int zmN sN).
Qed.

Lemma osum_of_nat_bis n X f: natp n -> (forall i, i <c n -> natp (X i)) ->
  inc f (permutations n) ->
  nds_sc n X f  =  osumf X n.
Proof.
move => nN ax /permutationsP[bf sf tf].
have ax2: forall i, i <c n -> natp (X (Vf f i)).
  move => i /(NltP nN) libe; apply: ax;apply/(NltP nN).
  move: (proj1 (proj1 bf)) => ff; Wtac.
by rewrite /nds_sc ! osum_of_nat // -{1} sf - tf csum_Cn'.
Qed.

Lemma nds_of_nat n X:  natp n -> (forall i, i <c n -> natp (X i)) ->
  nds_card n X = \1c.
Proof.
move => nN Xp.
suff: (nds_sums n X) = singleton (osumf X n).
  by rewrite /nds_card;move => ->; rewrite cardinal_set1.
apply: set1_pr1.
  move:(permutation_id n) => ip;  apply/funI_setne; ex_tac.
move => s/funI_P [z zp ->]; exact:osum_of_nat_bis.
Qed.


Lemma nds_same_deg_s2 e c1 c2 r1 r2:
  ordinalp e -> ordinalp c1 -> r1 <o oopow e -> 
  \0o <o c2 -> ordinalp r2 ->
  (oopow e *o c1 +o r1) +o (oopow e *o c2 +o r2) =
  (oopow e *o (c1 +o c2) +o r2).
Proof.
set p := oopow e;move => oe oc1 pa pb or2.
have op := OS_oopow oe.
have or1:= (proj31_1 pa).
have oc2:= (proj32_1 pb).
have oa := OS_prod2 op oc1.
have ob := OS_prod2 op oc2.
have oc := OS_sum2 ob or2.
have H:=(ord_negl_prod or1 op  pb (indecomp_prop1 (indecomp_prop4 oe) pa)).
rewrite - (osumA oa or1 oc) (osumA  or1 ob or2) H (osumA oa ob or2).
by rewrite - (oprodD oc1 oc2 op).
Qed.

Lemma nds_same_deg_sn e (c r: fterm) n:
  natp n -> \0c <c n -> ordinalp e -> 
  (forall i, i<c n -> \0o <o c i /\ r i <o oopow e) ->
  osumf (fun i => oopow e *o (c i) +o r i) n =
  oopow e *o (osumf c n) +o (r \0c).
Proof.
move => nN nz oe.
move: (OS_oopow oe) => op.
move: (cpred_pr nN (nesym (proj2 nz))) => [pN ->].
move: (cpred n) pN; clear n nN nz; apply: Nat_induction.
  rewrite succ_zero => H.
  move: (H _ (clt_01)) => [[[_ oc _] _] [[or _ _] _]].
  have ot:= (OS_sum2 (OS_prod2 op oc) or).
  by rewrite osum_f1 // (osum_f1 oc).
move => n nN Hrec ol.
move:(true_below_rec (NS_succ nN) ol) => [sa [/proj32_1 ocn rs]].
move:(proj31_1 (proj2(sa _ (succ_positive n)))) => or0.
rewrite (osum_fS _ (NS_succ nN)) (Hrec sa).
move:(true_below_rec nN sa) => [sb [[/proj32 oc0 /nesym cnz]_]].
have pc: \0o <o osumf c (csucc n).
  have /(OS_osumf nN) ha: ord_below_n c n by move => i /sb [/proj32_1 ].
  rewrite (osum_fS _ nN); apply: (ord_ne0_pos  (OS_sum2 oc0 ha)).
  by case/(osum2_zero oc0 ha).  
by rewrite (nds_same_deg_s2 oe ocn rs pc or0) (osum_fS _ (NS_succ nN)).
Qed.

Lemma nds_same_deg_s_perm e (c r: fterm) n f
      (X := (fun i => oopow e *o (c i) +o r i)): 
  natp n -> \0c <c n-> ordinalp e -> 
  (forall i, i<c n -> \0o <o c i /\ r i <o oopow e) ->
  (forall i, i<c n ->  c i <o omega0)  ->
  inc f (permutations n) ->
  nds_sc n X f =
  oopow e *o (osumf c n) +o (r (Vf f \0c)).
Proof.
move => nN nz oe h1 h2 fp.
move:(fp) =>  /Zo_P [] /functionsP [fz sz tz] bz.
have sa: forall i, i <c n -> \0o <o c (Vf f i) /\ r (Vf f i) <o oopow e.
  move => i /(NltP nN) lin; apply: h1; apply /(NltP nN); Wtac.
transitivity
 (osumf(fun i => oopow e *o (c (Vf f i)) +o r (Vf f i)) n).
  by apply:(osumf_exten nN) => i lin.
have h3: forall i, i <c n -> natp (c i).
  by move => i  /h2 /olt_omegaP. 
rewrite (nds_same_deg_sn nN nz oe sa).
by rewrite - /(nds_sc n c f) (osum_of_nat_bis nN h3).
Qed.

Lemma nds_card_same_deg_bd e X n: natp n -> \0c <c n ->
  (forall i, i <c n -> \0o <o X i) ->
  (forall i, i <c n -> odegree (X i) = e) ->
  nds_card n X <=c n.
Proof.
move => nN nz pa pb.
pose c i := CNF_lc (X i).
pose r i := CNF_rem (X i).
have pc i: i<c n -> [/\ \0o <o c i, c i <o omega0, (r i) <o oopow e
      & X i = oopow e *o (c i) +o (r i)].
  move => lein; move: (pa _ lein) (pb _ lein) => sa sb.
  by move: (the_CNF_split sa); rewrite sb. 
have oe: ordinalp e.
  rewrite -(pb _ nz); exact: (OS_degree (proj32_1 (pa _ nz))).
set s1:= oopow e *o (osumf c n).
have os1: ordinalp s1. 
  apply: (OS_prod2 (OS_oopow oe) (OS_osumf nN _)) => i lin.
  by move:(pc _ lin) => [ /proj32_1 h _ _ _].
set E := (permutations n).
set F1 := (fun_image E (nds_sc n X)).
set F2 := (fun_image E (fun f => r (Vf f \0c))).
have pe i: i<c n ->  c i <o omega0  by move =>  /pc[].
have pd i: i<c n -> \0o <o c i /\ r i <o oopow e by move => /pc[].
have pf: forall f, inc f E -> nds_sc n X f = s1 +o (r (Vf f \0c)).
  move => f fp; rewrite - (nds_same_deg_s_perm nN nz oe pd pe fp).
  by apply:(nds_sc_exten nN _ fp) => i /pc [].
have pg: forall z, inc z E -> inc (Vf z \0c) n.
  move => z /Zo_P [] /functionsP [fz sz tz] bz; Wtac.
  by rewrite sz; apply/(NltP nN). 
have osF: ordinal_set F2.
  by move=> s /funI_P [z /pg  /(NltP nN) /pc  [_ _ /proj31_1 h _] ->].
suff: nds_card n X = cardinal (fun_image n r).
  move ->; rewrite -{2} (card_nat nN); apply:fun_image_smaller.
symmetry.
apply/card_eqP.
exists (Lf (fun z => s1 +o z) F2 F1); split.
* apply:lf_bijective.
  + by move => s /funI_P [z ze -> ]; apply /funI_P; ex_tac; rewrite (pf _ ze).
  + by move => u v /osF ou /osF ov /(osum2_simpl ou ov os1).
  + move => s /funI_P [z ze -> ];  rewrite (pf _ ze).
    exists (r (Vf z \0c)) => //;apply/funI_P; ex_tac.
* rewrite lf_source;set_extens t.  
  move => /funI_P [z /pg ze -> ];  apply /funI_P; ex_tac.
  move => /funI_P [i /(NltP nN) iI ->]; apply /funI_P.
  move: (permutation_exists1 nN iI) => [f fe <-]; ex_tac.
* aw.
Qed.


Lemma osum_negl_rec n e X: ordinalp e -> natp n ->
  (forall i, i<c n -> X (csucc i) <o oopow e) -> X \0c = oopow e ->
  osumf X (csucc n) = oopow e.
Proof.
move => oe nN os xz; move: (OS_oopow oe) => oope.
have obl: ord_below_n X (csucc n).
   move => i lin; case: (equal_or_not i \0c) => iz; first by rewrite iz xz.        move: (cpred_pr (NS_lt_nat lin (NS_succ nN)) iz) => [pN pv].
   have: cpred i <c n by apply/(cleSltP pN); rewrite - pv;apply/cltSleP.
   by move/os /proj31_1; rewrite - pv.
rewrite (osum_f1r nN obl) xz {obl}.
apply:   (ord_negl_opow _ oe).
move: n nN os; apply: Nat_induction.
  by rewrite osum_f0 => _; apply: oopow_pos.
move => n nN Hrec ax. move:(true_below_rec nN ax)=> [ax1 ax2].
rewrite (osum_fS _ nN);move: (Hrec ax1); set v := osumf _ _ => lvo.
exact: (indecomp_prop2 ax2 lvo (indecomp_prop4 oe)).
Qed.
  

Lemma nds_card_different_deg n: natp n ->
  nds_ax oopow (csucc n) /\ \2c ^c n <=c  nds_card (csucc n) oopow. 
Proof.
move => nN; move: (NS_succ nN) => sN.
have ax: nds_ax oopow (csucc n).
   by move => i /oclt/proj31_1/OS_oopow.
split => //.
rewrite - card_setP; set E := \Po n; set F := (\Po (csucc n)).
pose p  K s := let k := cardinal K in
  [/\ inc s (permutations (csucc n)),
    (forall x, inc x K -> exists2 i, i <c k & x = (Vf s i)),
    (forall i, i <c k -> inc (Vf s i) K),
    Vf s k = n &
    (forall i j, i<c j -> j <c k -> (Vf s i) <c (Vf s j))].
have oopn : ordinalp (oopow n) by apply: (ax _ (cltS nN)).

have Hb K s: inc K E -> p  K s ->
   let s2 := osumf (fun z => oopow(Vf s z)) (cardinal K) in 
    ordinalp s2 /\ nds_sc (csucc n) oopow s = oopow n +o s2.
  move => ha [hb hc hd he hf]; rewrite /nds_sc.
  move/setP_P: ha => Kb.
  set k := cardinal K.
  move:(sub_smaller Kb); rewrite (card_nat nN) -/k => lkn.
  set s2 :=  osumf (fun z => oopow (Vf s z)) k.
  have kN:= (NS_le_nat lkn nN).
  move: (cdiff_pr lkn) (NS_diff k nN); set m := n -c k => mp mN.
  have eq1 : k +c (csucc (n -c k)) = csucc n by rewrite (csum_nS _ mN) mp.
  move:(nds_ax_perm sN ax hb); rewrite - eq1; move=> [sax oo1].
  rewrite(osum_fA kN  (NS_succ mN) sax) -/s2.
  have sbd i: i <c (csucc n) -> Vf s i <c (csucc n).
     move/permutationsP: hb => [[[fp _] _] ss ts].
     move/(NltP sN) => ins; apply(NltP sN); Wtac.
  split; last (apply: f_equal2; last exact).
    apply: (OS_osumf kN) => i li; apply: ax; apply: sbd; apply: (clt_ltT li).
    by apply/(cltSleP nN).
  have he':oopow (Vf s (\0c +c k)) = oopow n by rewrite (csum0l (CS_nat kN)) he.
  apply: (osum_negl_rec (OS_nat nN) mN _ he').
  move => i lim; apply: opow_Mo_lt; apply:oclt.
  rewrite (csum_Sn _ (NS_lt_nat lim mN)).
  have pa: (csucc (i +c k))  <c csucc n.
    by rewrite  - mp csumC; apply/cltSS;apply(csum_Meqlt kN). 
  move: (sbd _ pa) => /(cltSleP nN) /cle_eqVlt; case => // vn.
  move/permutationsP: hb => [[[fp injf] _] ss ts].
  have x1: inc (csucc (i +c k)) (source s) by rewrite ss; apply/NltP.
  have x2: inc k (source s) by rewrite ss; apply/(NltP sN); apply/cltSleP.
  rewrite - he in vn.
  case: (proj2 (clt_leT (cltS kN) (cleSS (Nsum_M0le i kN)))).
  by rewrite csumC  (injf _ _ x1 x2 vn).
have Hc K1 K2 s1 s2: inc K1 E -> p K1 s1 -> inc K2 E -> p  K2 s2 ->
   nds_sc (csucc n) oopow s1 =  nds_sc (csucc n) oopow s2 -> K1 = K2.
  move => hk1 hs1 hk2 hs2; move:  (Hb _ _ hk1 hs1) => [os1 ->].
  move: (Hb _ _ hk2 hs2) => [os2 ->].
  move / (osum2_simpl os1 os2 oopn).
  pose q n K f :=
    [/\ sub K n, (forall x, inc x K -> exists2 i, i <c cardinal K & x = f i), 
     (forall i, i <c cardinal K -> inc (f i) K) &
     (forall i j, i <c j -> j <c cardinal K -> f i <c f j)]. 
  have: q n K2 (Vf s2).
    by move/setP_P: hk2 => pe; move: hs2 => [pa pb pc pd]; split.
  have: q n K1 (Vf s1).
    by move/setP_P: hk1 => pe; move: hs1 => [pa pb pc pd]; split.
  set f1 := Vf s1; set f2 := Vf s2; move: f1 f2.
  clear os1 os2 sN hs1 hs2 hk1 hk2  Hb E F oopn ax p s1 s2;  move: n nN K1 K2.
  have prop0 n K: natp n ->  sub K (csucc n) ->  ~inc n K -> sub K n.
    move => nN ha h t tK.
    move: (ha _ tK)=> /(NltP (NS_succ nN)) /(cltSleP nN) /cle_eqVlt; case.
      by move => tn; case: h; rewrite - tn.
    by move/(NltP nN).
  have prop1 n K : natp n -> sub K n -> natp (cardinal K).
    by move => nN /sub_smaller; rewrite (card_nat nN) => x; apply (NS_le_nat x).  have hw n K f: natp n -> sub K n -> (forall i, i <c cardinal K -> inc (f i) K)
    -> osumf (fun z => oopow (f z)) (cardinal K) <o oopow n.
    move => nN ha hb.
    have ii: indecomposable (oopow n) by apply: (indecomp_prop4 (OS_nat nN)).
    have: forall i, i <c (cardinal K) ->  (f i) <o n.
       by move => i /hb /ha /(NltP nN)/oclt.
    move: (prop1 _ _ nN  ha).
    move:(cardinal K); apply:Nat_induction.
       by move => _; rewrite osum_f0; apply: (oopow_pos (OS_nat nN)).
    move => m mN Hr ax; move: (true_below_rec mN ax) => [ax1 /opow_Mo_lt ax2].
    rewrite (osum_fS _ mN); exact:(indecomp_prop2 ax2 (Hr ax1) ii).
  have prop3 n K f: let K1 := K -s1 n in let k1 := cpred (cardinal K) in 
      natp n -> q (csucc n) K f -> 
      inc n K -> [/\ [/\ cardinal K  = csucc (cardinal K1),
                      natp (cardinal K1) & k1 = cardinal K1],
                  k1 <c cardinal K, f k1 = n ,
       forall i, i <c cardinal K1 -> inc (f i) K1 &  q n (K -s1 n) f].
    move => /= nN [qa qb qc qd] nK; move: (NS_succ nN) => snN.
    move: (cpred_pr5 nK) (csucc_pr2 nK).
    have Kn:=  (prop1 _ _ snN qa).
    set k := cpred _; set k' := cardinal _ => eq1 eq2. 
    have ckz: cardinal K <> \0c by  rewrite eq2; apply: succ_nz.
    move: (cpred_pr Kn ckz). rewrite - eq1 -/k; move => [kN eq3].
    move: (cltS kN); rewrite - eq3 eq1 => lt1.
    move: (qb _ nK) => [i lik fv]; move: lik; rewrite {1} eq3  =>/(cltSleP kN).
    rewrite {1} eq1 ;move/cle_eqVlt; case => ik1; last first.
      move:  (qd _ _ ik1 lt1)  (qc _ lt1); rewrite - fv => pd /qa /(NltP  snN).
      by move/(cltSleP nN) /cleNgt; case.
    rewrite ik1 in fv; clear i ik1.
    move: (kN); rewrite eq1 => kN'.
    have hu:  forall i, i <c k' -> inc (f i) (K -s1 n).
      move => i lik; apply/setC1_P.
      move:(qd _ _ lik lt1) (qc _ (clt_ltT lik lt1)) => pd jsk.
      by split => // eq4; case: (proj2 pd); rewrite eq4 fv. 
    split => //;  split.
    - by apply:(prop0 _ _ nN); [ move => t/setC1_P [/qa] | move=> /setC1_P [_]].
    - move => x /setC1_P [/qb [i lik ->] fjn]; exists i => //.
      move: lik; rewrite eq2 ; move/(cltSleP kN'); case/cle_eqVlt => //.
      by move => jk; case: fjn; rewrite jk  fv.
    - done.
    - move => i j lij ljk; apply:(qd _  _ lij (clt_ltT ljk lt1)).
  apply: Nat_induction.
    by move => K K' f f'' [ /sub_set0 -> _ _ _] [/sub_set0 -> _ _ _] _.
  move => n nN Hrec K K' f f' qp qp'.
  have sN := NS_succ nN.
  have pa t: inc t  (csucc n) -> inc t n \/ t = n.
    move/(NltP sN) /(cltSleP nN) /cle_eqVlt; case; first by right.
    by move/(NltP nN); left.
  move: (qp) (qp')=> [pr1 pr2 pr3 pr4] [pr1' pr2' pr3' pr4'].
  move: (prop0 _ _ nN pr1)  (prop0 _ _ nN pr1') => hu hv.
  have cKN:= prop1 _ _ sN pr1; have cKN':= prop1 _ _ sN pr1'.
  have hx: ord_below_n (fun z => oopow (f z)) (cardinal K).
       by move => i /pr3 /pr1 /(NS_inc_nat sN) /OS_nat/OS_oopow.
  have hx': ord_below_n (fun z => oopow (f' z)) (cardinal K').
       by move => i /pr3' /pr1' /(NS_inc_nat sN) /OS_nat/OS_oopow.
  case: (inc_or_not n K) => inK; case: (inc_or_not n K') => inK'.
  -     move: (prop3 _ _ _  nN  qp inK).
    set K1 := K -s1 n; set k1 := cpred _. 
    move => [[k1v k1N k2v] lt1 fk1 fk2 qp1].
    move: (prop3 _ _ _  nN  qp' inK').
    set K1' := K' -s1 n; set k1' := cpred _.
    move => [[k1v' k1N' k2v'] lt1' fk1' fk2' qp1'].
    have ax1: ordinalp (osumf (fun z => oopow (f z)) k1).
      apply: (OS_osumf (NS_lt_nat lt1 cKN)) => i ii.
      apply: (hx _ (clt_ltT ii lt1)).
    have ax1': ordinalp (osumf (fun z => oopow (f' z)) k1').
      apply: (OS_osumf (NS_lt_nat lt1' cKN')) => i ii.
      apply: (hx' _ (clt_ltT ii lt1')).
   have o3: ordinalp (oopow n) by apply/(OS_oopow (OS_nat nN)).
   rewrite k1v k1v' (osum_fS _ k1N)  (osum_fS _ k1N') - k2v fk1 - k2v' fk1'.
   move/(osum2_simpl ax1 ax1' o3); rewrite  k2v k2v' => ss.
   by rewrite - (setC1_K inK) - (setC1_K inK') -/K1 (Hrec _ _ _ _ qp1 qp1' ss).
  - move: (hv inK') => ra eq0.
    move: (hw _ _ _ nN ra pr3'); rewrite - eq0 => lt1.
    move: (prop3 _ _ _  nN  qp inK) => [_ qb qc qd].
    by case: (oleNgt (osum_Mon1 cKN qb hx)); rewrite qc.  
  - move: (hu inK) => ra eq0.
    move: (hw _ _ _ nN ra pr3); rewrite  eq0 => lt1.
    move: (prop3 _ _ _  nN qp' inK') => [qa qb qc qd].
    by case: (oleNgt (osum_Mon1 cKN' qb hx')); rewrite qc.  
  - by move: (hu inK) (hv inK')  => ra rb; apply: Hrec.
    rewrite /nds_card.
pose f K := (Lf (nth_elt_dbl (K +s1 n) (csucc n)) (csucc n) (csucc  n)).
have fp K: inc K E -> p K (f K) by apply:nth_elt_dbl_prop_bis.
pose g K := nds_sc (csucc n) oopow (f K).
have: {inc E &, injective g}.
   move => k1 k2 k1E k2E; exact: (Hc _ _ _ _ k1E (fp _ k1E) k2E (fp _ k2E)).
move/(cardinal_fun_image) => <-; apply: sub_smaller.
move=> y/funI_P [K kE ->]; move:(fp _ kE) => [xx _ _ _ _]; apply/funI_P; ex_tac.
Qed.

Lemma nds_sd1 n: natp n -> \2c ^c n <=c nds_F (csucc n).
Proof.
move => nN.
move: (nds_card_different_deg nN) => [ra rb].
move: (nds_f_def (NS_succ nN)) => [_ _ _ rc].
exact: (cleT rb  (rc _ ra)).
Qed.


Lemma nds_sd2 n: natp n -> \2c <c n -> n <c (nds_F n). 
Proof.
move => nN ngt2.
have nnz:= nesym (proj2(clt_ltT clt_02 ngt2)).
move:(cpred_pr nN nnz) => [n1N sa].
have npz: cpred n <> \0c.
  move => h; move: sa (proj1 ngt2). 
  by rewrite h succ_zero => -> /(cltNge clt_12).
move:(cpred_pr n1N npz) => [n2N sb].
have pN:= NS_pow NS2 n2N.
have ha : \2c <=c \2c ^c cpred (cpred n).
   apply:(cpow_Mle1 CS2) => mz.
   by case:(proj2 ngt2); rewrite sa sb mz succ_zero succ_one.
have hb: csucc (cpred n) <c \2c ^c cpred n.
  rewrite sb (cpow_succ _ n2N) cprodC two_times_n.
  rewrite (Nsucc_rw (NS_succ n2N)) (Nsucc_rw n2N) - csumA card_two_pr csumC.
  exact:(csum_Mlelt pN  ha (cantor (CS_nat n2N))).
rewrite sa; exact: (clt_leT hb (nds_sd1 n1N)).
Qed.



Lemma nds_type0 n X: 
  natp n -> \2c <=c n -> nds_ax X n -> (exists2 i, i<c n & X i = \0o) -> 
  exists2 Y, nds_ax Y n & nds_card n X <c nds_card n Y.
Proof.
move => nN ln2 ax Hx.
have nz:= (card_ge2_nz ln2).
move: (cpred_pr nN nz); set m := cpred n; move => [mN nv].
have: exists Z,
    [/\ nds_ax Z n, nds_card n X = nds_card n Z & Z m = \0o].
  move: Hx => [i lin Xz].
  have iin: inc i n by apply /(NltP nN).
  move /(NltP nN): (cpred_lt nN nz) => ipn.
  move:(transposition_prop nN iin ipn).
  set G := Lf _ _ _; move => [Gp Gi Gpn Go GGx].
  move:(proj1(nds_ax_perm nN ax Gp));set Z := (fun z => X (Vf G z)) => axZ.
  by exists Z; rewrite (nds_card_perm_inv nN ax Gp) /m /Z Gpn. 
move => [Z [axZ -> Z0]]; clear X ax Hx.
wlog xnz: Z Z0 axZ/ exists2 i, i <c m & Z i <> \0o. 
  move=> Ha; case: (p_or_not_p(exists2 i, i <c m & Z i <> \0o)) => Hx.
    by apply: Ha.
  have Xn: forall i, i <c n -> natp (Z i).
    rewrite nv => i /(cltSleP mN)/ cle_eqVlt; case.
      move ->;erewrite Z0; apply:NS0.
    move => im; case: (equal_or_not (Z i) \0o) => h; first by rewrite h; fprops.
    by case: Hx; exists i.
  pose X1 := fun i => Yo (i = m) \0o \1o.
  have X1n: forall i, i <c n -> natp (X1 i).
     rewrite /X1; move => i ii; Ytac h; fprops.
  have X1m:  X1 m = \0o by rewrite /X1; Ytac0.
  move: ln2; rewrite - succ_one {1}nv;move/(cleSSP CS1 (CS_nat mN))/cge1P => n1.
  have X10: exists2 i, i <c m & X1 i <> \0o.
    exists \0c=> //; rewrite /X1 (Y_false (proj2 n1)); fprops.
  have ax1: nds_ax X1 n by move => i /X1n /OS_nat.
  by rewrite (nds_of_nat nN Xn) -(nds_of_nat nN X1n); apply: Ha.
have lmn:  m <c n by move: (cltS mN); rewrite nv.
have axm: nds_ax Z m by move => i lin; apply: axZ; move: (clt_ltT lin lmn).
have alsp: forall v, inc v (nds_sums m Z) -> \0o <o v.
  move => v /funI_P [s sp ->].
  move: (nds_ax_perm mN axm sp) => [ha hb].
  apply:(osum_fnz mN ha).
  move/permutationsP: sp => bsp.
  move: (inverse_bij_bp bsp) => [[[isb _] _] iss ist].
  move: bsp=> [bs ss ts].
  move: xnz => [k lkm Zk].
  have ktf: inc k (target s) by  rewrite ts; apply/(NltP mN).
  move: (inverse_V bs ktf); rewrite -/k => kv.
  exists  (Vf (inverse_fun s) k); last by rewrite kv.
  by apply/(NltP mN); Wtac; rewrite iss - ts.
have [e oe ep]: exists2 e, ordinalp e &  
  (forall i, i<c m -> Z i <o oopow e). 
  suff: exists2 x, ordinalp x & (forall i, i<c m -> Z i <=o x).
    move => [x ox xle]; move:(OS_succ ox) => osx; exists (osucc x) => //.
    move => i /xle le; apply: (ole_ltT le); apply/oleSltP.
    by apply:ord_eps_p1.
  clear nv Z0 xnz lmn alsp.
  move: m mN axm; apply:Nat_induction.
    move => _; exists \0o; [ exact: OS0 | by  move => i /clt0].
  move => m mN Hrec h. move: (true_below_rec mN h)=> [sa sb].
  move: (Hrec  sa) => [a oa ap].
  move:(omax_p1 oa sb) => [sc sd]; exists (omax a (Z m)).
      exact (proj32 sc).
  move => i /(cltSleP mN) /cle_eqVlt; case; first by move ->.
  move/ap => le1; apply: (oleT le1 sc).
pose T u i := Yo (i = m) u (Z i).
have To u: ordinalp u -> nds_ax (T u) n.
   rewrite nv /T => ou i /(cltSleP mN) /cle_eqVlt; case => li; try Ytac0 => //.
   by rewrite (Y_false (proj2 li)); apply: axm.
have Tuv u: ordinalp u -> 
  forall v, inc v (nds_sums m Z) -> inc (u +o v)  (nds_sums n (T u)).
  move => ou v /funI_P [s sp ->].
  move: (extension_perm mN sp).
  set es := extension s m m; move => [esp esout esin].
  rewrite - nv in esp; apply/funI_P; ex_tac.
  rewrite/nds_sc; rewrite nv (osum_fS _ mN) esout/T; Ytac0; congr osum2.
  apply: (osumf_exten mN) => i /(NltP mN) lin; rewrite esin //.
  have: inc (Vf s i) m by move/permutationsP: sp => [[[ff _] _ ] sf tf]; Wtac.
  by move/(NltP mN) => /proj2 nn; Ytac0.
set u := oopow e.
have ou: ordinalp u by exact:(OS_oopow oe).
have us: inc u (nds_sums n (T u)).
  apply/funI_P; move: (rotation_prop mN); rewrite - nv.
  set f := Lf _ _ _; set g := Lf _ _ _.
  move => [pf pg ifg [fo fa1 fa2 gn ga]].
  exists f; first exact.
  rewrite /T /nds_sc nv (osum_negl_rec oe mN) //.
    by move => i lin; rewrite (fa1 _ lin)  (Y_false (proj2 lin)); apply: ep.
  by rewrite fo; Ytac0.
move: (To  u ou) => axT.
exists (T u); first exact: (To  u ou).
apply/(cleSltP (NS_nds_card Z nN)).
have ->: nds_card n Z = cardinal(nds_sums m Z).
  rewrite /nds_card /nds_sums; congr cardinal.
  set_extens z; last first.
    rewrite nv; move:  (Tuv _ OS0) => rb zv.
    move: (rb _ zv); rewrite (osum0l(nds_sums_oset mN axm zv)) - nv.
    congr inc;apply:(nds_sums_exten nN)=> i lin; rewrite /T.
    by Ytac h => //; rewrite h Z0.
  move => /funI_P [f fp ->].
  rewrite nv in fp; move: (partial_rotation mN fp).
  set k := Vf (inverse_fun f) m.
  set (g:= fun i => Yo (i <c k) (Vf f i) (Vf f (csucc i))).
  move => [lekn fk axg gp].
  apply/funI_P; ex_tac.
  move: (cdiff_pr lekn) (NS_diff k mN); set q:= _ -c _ => sb qN.
  move/(cleSSP (proj31 lekn) (proj32 lekn)): (lekn) => li0n1.
  have kN:=(NS_le_nat lekn mN).
  have skN:=(NS_succ kN).
  pose g0 i := Z (Vf (Lf g m m) i).
  pose f0 i := Z (Vf f i).
  move/permutationsP: (fp) => [bf sf tf]; move: (proj1(proj1 bf)) => ff.
  rewrite -nv in sf tf.
  have ax1: ord_below_n f0 (csucc k +c q).
    rewrite (csum_Sn _ kN) sb - nv => i lin; rewrite /f0; apply: axZ.    
    by apply/(NltP nN);  Wtac; rewrite sf; apply/(NltP nN).
  have ax2: (ord_below_n g0 (k +c q)).
    rewrite sb => i/(NltP mN) lin; rewrite /g0; aw; apply axm. 
    by move: (axg _ lin) => /(NltP mN) h1. 
  have osa: ordinalp (osumf f0 k).
    apply: (OS_osumf kN) => i lin; rewrite /f0; apply: axZ.
    apply/(NltP nN); Wtac; rewrite sf;apply/(NltP nN).
    exact: (clt_ltT lin (cle_ltT lekn lmn)).
  move:(osum_fA skN qN ax1);rewrite (csum_Sn _ kN) sb nv -/(nds_sc _ _ _) => ->.
  rewrite (osum_fS _ kN) {2} /f0 fk Z0 (osum0l osa).
  rewrite /nds_sc - {3} sb (osum_fA kN qN ax2); apply: f_equal2.
    apply:(osumf_exten qN) => i lim; rewrite /f0/g0 /g.
    have ha: inc (i +c k) m. 
     apply /(NltP mN); rewrite - {1} sb (csumC _ q); apply:(csum_Mlteq kN lim).
    have hb: ~(i +c k <c k).
      move:(csum_Mle0 i (proj31 lekn)) => ua ub. 
      case: (cleNgt ua ub).
    by aw; Ytac0; rewrite (csum_nS _ kN).
  apply:(osumf_exten kN) => i lim; rewrite /f0/g0 /g. 
  aw; [ by Ytac0 |apply /(NltP mN); apply: (clt_leT lim lekn) ].
have alsp': ~inc \0o (nds_sums m Z) by move /alsp; case.
rewrite - (csucc_pr alsp').
have: injection( Lf (osum2 u) (nds_sums m Z +s1 \0o) (nds_sums n (T u))).
  have H x: inc x (nds_sums m Z +s1 \0o) -> ordinalp x.
    case /setU1_P; last by move ->; apply: OS0.
    move => /funI_P [p pp ->]; exact:proj2 (nds_ax_perm mN axm pp).
  apply: lf_injective.
    by move => x/setU1_P; case; [ apply: Tuv | move ->; rewrite (osum0r ou)].
  move => x y /H xs /H ys; apply: (osum2_simpl xs ys ou).
move /incr_fun_morph; aw.
Qed.



Definition nds_type X n k :=
  (forall i,  i<c n ->  \0o <o X i) /\
  (exists m, [/\ ordinalp m,
     (forall i,  i<c n -> m <=o odegree (X i)) &
     cardinal (Zo n (fun i => odegree (X i) = m)) = k]).

Definition nds_FA n k:= 
 \osup (Zo Nat (fun z => exists X, [/\ nds_ax X n, nds_card n X = z
     & nds_type X n k])).


Lemma nds_type_exists_bd n k:
   natp n -> \0c <c k -> k <=c n -> exists X,
   [/\ nds_ax X n, nds_rel_small_ax \1c \1c X n  & nds_type X n k].
Proof.
move => nN kp lkn. 
have kz := (nesym (proj2 kp)).
pose X i := Yo (i <c k) \1o omega0.
have ha: nds_ax X n  by move => t _; rewrite /X; Ytac h; fprops.
have hc: nds_rel_small_ax \1c \1c X n.
  move => t _; rewrite /X; Ytac h; last exact:rel_small_omega.
  move:rel_small_one; apply: (rel_small_le1 ole_01).
have hb:forall i, i <c n -> \0o <o X i.
  move => i _; rewrite /X; Ytac h; [apply: olt_01| apply: olt_0omega].
exists X; split => //; split => //. 
move: odegree_one odegree_omega => dg1 dgo.
exists \0o; split; [ exact: OS0 | by move => i /ha/OS_degree /ole0x | ].
set E := Zo _ _. 
move:(NS_le_nat lkn nN) => kN.
rewrite -(card_nat kN); congr cardinal.
set_extens t.
  move => /Zo_P [ /(NltP nN) lth]; rewrite /X; Ytac h1.
   by move => _; apply/(NltP kN).
  by rewrite dgo => ba; case: card1_nz.
move => /(NltP kN) ltl;apply: Zo_i.
  apply /(NltP nN); exact:(clt_leT ltl lkn).
by rewrite /X; Ytac0; rewrite dg1.
Qed.


Lemma nds_FA_def n k (v := nds_FA n k):
  natp n -> \0c <c k -> k <=c n -> 
  [/\ natp v,
   v <=c  (nds_F n),
   (forall X,  nds_ax X n -> nds_type X n k -> nds_card n X <=c v) &
   (exists  X, [/\ nds_ax X n, nds_type X n k & nds_card n X = v])].
Proof.
move=> nN kp kn.
rewrite /v /nds_FA; set T:= Zo Nat _.
have ta: sub T Nat by apply: Zo_S.
have tb: natp (nds_F n)  by move: (nds_f_def nN) => [ok _].
have tc: (forall i, inc i T -> i <=c  (nds_F n)).
  move => i /Zo_P [_ [X [ax <- _]]].
  by move: (nds_f_def nN) => [_ _ _ p4]; apply: p4.
move: (Nat_sup_pr ta tb tc) => [pa pb pc pd].
have pe X: nds_ax X n ->  nds_type X n k -> inc (nds_card n X) T.
  by move => ax ax2; apply: Zo_i;[ apply: NS_nds_card | exists X].
split => //.
  by move=> X ax ax2; apply: pc; apply: pe.
case: pd; last by move /Zo_P=> [_ [X [p1 p2 p3]]];exists X; split.
move: (nds_type_exists_bd nN kp kn)=> [X [ax _ tx]] te.
empty_tac1 (nds_card n X);apply:pe.
Qed.

Lemma nds_FA_21: nds_FA \2c \1c = \2c.
Proof.
move: (nds_FA_def NS2 clt_01 cle_12);rewrite nds_f2.
set v := (nds_FA \2c \1c);move => [vN lv2 ha hb].
apply: (cleA lv2).
move:(nds_card2); set X := variant _ _ _; move => [Xa <-].
have Ha:  forall i, i <c \2c -> \0o <o X i.
  move => i _; rewrite /X /variant; Ytac h; [ apply:olt_01 | apply:olt_0omega].
have Hb: forall i, i <c \2c -> \0o <=o odegree (X i).
  by move => i /Ha /proj32_1 /OS_degree /ole0x.
apply: (ha X Xa);  split => //.
 exists \0o; split => //; first  exact: OS0.
apply/set_of_card_oneP; exists \0o; apply: set1_pr.
  by apply: (Zo_i inc_C0_C2);rewrite /X variant_true1 odegree_one.
move => t /Zo_P [/C2_P]; case => -> //.
by rewrite /X variant_false1 odegree_omega.
Qed.


Lemma nds_F_FA_def n (g := nds_FA n) (f:= nds_F n):
  natp n -> \2c <=c n ->
  (forall k,  \0c <c k -> k <c n  -> (g k) <=c f) /\ 
  (exists k, [/\ \0c <c k,  k <c n & (g k) = f]).
Proof.
move=> nN nge2.
split.
  by move => k kp kn ; move: (nds_FA_def nN kp (proj1 kn)) => [_ pa _].
case: (equal_or_not n \2c) => en2.
  move: clt_12 clt_01 => ra rb. 
  rewrite /f /g en2 nds_f2; exists \1c; split => //; exact: nds_FA_21.
move: (nds_f_def nN); rewrite -/f; move=> [_  _ [X ax fv] pd].
move: (clt_leT clt_02 nge2) => np.
case: (p_or_not_p (exists2 i, i<c n & X i = \0o)) => tp0.
  move:(nds_type0 nN nge2 ax tp0) => [y ay xsy].
  by case: (cleNgt(pd y ay)); rewrite - fv.
set E:= fun_image  n (fun i => odegree (X i)).
set m := intersection E.
set F :=  (Zo n (fun i => odegree (X i) = m)).
have s1: sub F n by apply Zo_S.
move: (sub_smaller s1); rewrite (card_nat nN) => kn.
have osE: ordinal_set E.
  rewrite /E; move=> t /funI_P [z /(NltP nN) /ax zi ->]. 
  by apply: OS_degree.
have neE: nonempty E by apply:funI_setne; exists \0c; apply/(NltP nN).
move: (ordinal_setI neE osE); rewrite -/m => mE.
set k := cardinal F.
have kp: \0c <c k.
  move: mE => /funI_P [z zn zv].
  apply: (card_ne0_pos (CS_cardinal _) (card_nonempty1 _)).
  by exists z; apply: Zo_i.
have kt: nds_type X n k.
  split.
    move => i lin; case: (ozero_dichot (ax _ lin)) => //.
    by move => baa;case: tp0; exists i.
  have om: ordinalp m  by apply:osE.
  exists m; split => // i /(NltP nN) lin.
  have uE: inc (odegree (X i)) E by apply /funI_P; ex_tac.
  by split; [ exact |  apply: osE | apply:setI_s1].
have eq1: g k = f.
  move:(nds_FA_def nN kp kn) => [_ _ qa  [Y [ax' x3 x4]]];  apply cleA.
    by rewrite /g - x4; apply: pd.
  by move: (qa _ ax kt); rewrite fv. 
case: (equal_or_not k n) => ekn; last by exists k.
case: (cltNge (nds_sd2 nN (conj nge2 (nesym en2)))); rewrite -/f - eq1 /g ekn.
move: (nds_FA_def nN np (cleR (CS_nat nN))).
move => [vB vle _ [Y [ay [ty1 [m' [om mp1 mp2]]] <-]]].  
have fsI:= (finite_set_nat nN).
move: (cardinal_setC4 (@Zo_S n (fun i => odegree (Y i) = m'))  fsI).
rewrite (card_nat nN) mp2  cdiff_nn => /card_nonempty /empty_setC si.
apply: (nds_card_same_deg_bd nN np ty1 (e:=m')).
by move => i  /(NltP nN) /si /Zo_hi.
Qed.


Definition nds_sc_K X K :=  osumf (fun z => X (nth_elt K z)) (cardinal K). 

Lemma nds_tn_of_nat n c K: natp  n ->
   (forall i, i<c n -> natp (c i)) ->
   inc K (\Po n) ->
   nds_sc_K c K = csumb K c.
Proof.
move => nN cIb /setP_P Kb. 
rewrite /nds_sc_K.
have KN: sub K Nat by move => t tK; move:(NS_inc_nat nN (Kb _ tK)).
set q := cardinal K; set f := nth_elt K.
have qN: natp q by  apply:(sub_nat_finite nN Kb).
have ha: forall i, i <c q -> inc (f i) K.
  move => i iq; apply:(nth_elt_inK (NS_lt_nat iq qN) KN iq).
have: forall j, inc j K -> exists2 i, i <c q & j = f i.
  move => j jK; apply: (nth_elt_surj KN qN jK).
have : forall i j, i <c q -> j <c q -> f i = f j -> i = j.
  move => i j /(NltP qN) iq /(NltP qN) jq.
  move:(proj2 (proj1 (proj1(nth_elt_bf KN qN)))).
  have ax: lf_axiom (nth_elt K) (cardinal K) K.
    by move => t /(NltP qN) /ha.
  aw => h; move: (h i j iq jq);aw.
have: forall i, inc i K -> natp (c i) by move => i /Kb /(NltP nN) /cIb.
have: finite_set K by apply/NatP.
move: q qN (f) (K) ha;  clear.
apply: Nat_induction.
  move => k K  _ _ _ _ sf.
  have h: domain (Lg K c) = emptyset.
    by bw; apply /set0_P => t /sf [i /clt0].
  by rewrite osum_f0 /csumb (csum_trivial h).
move => n nN Hrec f K fK fsK cN fi fs.
rewrite (osum_fS _ nN).
have nn:=  (cltS nN).
move:(fK _ nn) => fnK.
move: (setC1_K fnK); set K1:= K -s1 (f n) => Kv.
have h1: forall i, i <c n -> inc (f i) K1.
  move => i lin;  move:(clt_ltT lin nn) => h; move: (fK _ h).
  rewrite - Kv =>/setU1_P; case => // eq. move:(fi _ _ h nn eq) => ein.
  by case: (proj2 lin).
have h2:=(sub_finite_set (@sub_setC K (singleton (f n))) fsK).
have h3: forall i, inc i K1 -> inc (c i) Nat by move => i /setC1_P [/cN].
have h4: forall i j, i <c n -> j <c n -> f i = f j -> i = j.
  move => i j lin ljn;apply (fi _ _ (clt_ltT lin nn) (clt_ltT ljn nn)).
have h5:forall j, inc j K1 -> exists2 i : Set, i <c n & j = f i.
  move => j /setC1_P [/fs [i /(cltSleP nN) lin ->] h].
  case: (equal_or_not i n) => ein; first by case: h; rewrite ein.
  by exists i.
have fnk1:~ inc (f n) K1 by move/setC1_P=>  [ _ ].
have skN: natp (csumb K1 c) .
  apply:finite_sum_finite; hnf; rewrite/allf;bw;split => //.
  by move => t tk1; bw; apply: h3.
rewrite (Hrec f K1 h1 h2 h3 h4 h5) - Kv (csumA_setU1 _ fnk1).
by rewrite (osum2_2int (cN _ fnK) skN) csumC.
Qed.


Lemma nds_C_prop n a: natp n -> inc a n ->
  cardinal (\Po (n -s1 a)) =  \2c ^c (cpred n).
Proof.
move => mN an.
case: (equal_or_not n \0c) => mp; first by  move: an; rewrite mp => /in_set0. 
move:(cpred_pr mN mp) => [pnN pnv].
have ->: cpred n = cardinal (n -s1 a).
  by apply:succ_injective1;fprops;rewrite - pnv - (csucc_pr2 an) (card_nat mN).
by rewrite  cpowcr - card_setP.
Qed.


Definition nds_sc_Ktau X K tau :=
  nds_sc (cardinal K) (fun z => X (nth_elt K z)) tau.

Definition nds_tn_S X n:=
   unionf (\Po n) (fun K =>
      fun_image (permutations (cardinal K)) (nds_sc_Ktau X K)).
Definition nds_tn_C X n :=  cardinal (nds_tn_S X n).

Definition nds_tn_ax X n :=
    (forall i, i<c n -> \0o <o X i) /\
    exists2 e, ordinalp e & forall i, i<c n -> odegree (X i)  = e.

Definition nds_tn_sup n :=
  \osup(Zo Nat (fun z => exists2 X, nds_tn_ax X n & nds_tn_C X n = z)).

Lemma nds_tn_small X n:
   natp n -> nds_ax X n ->
   nds_tn_C X n <=c (\2c ^c n) *c (factorial n).
Proof.
move => nN ax.
rewrite /nds_tn_C /nds_tn_S; set F:= (fun K : Set => _).
apply: (cleT (csum_pr1_bis (\Po n) F)).
rewrite - card_setP cprod2cl cprodC  - csum_of_same.
apply:csum_increasing; bw => K KN; bw.
move/setP_P:KN => KN.
move:(sub_smaller KN); rewrite (card_nat nN) => ka.
have nKN:=(NS_le_nat ka nN).
move: (factorial_monotone nN ka); apply: cleT.
have pa: finite_set (cardinal K) by apply/NatP; rewrite double_cardinal. 
rewrite - (double_cardinal K) -(number_of_permutations pa).
by apply:fun_image_smaller.
Qed.

Lemma nds_tn_def n (f:=  nds_tn_sup n): natp n ->
  [/\ natp f, \0c <c f, 
  (exists2 X, nds_tn_ax X n & nds_tn_C X n = f) &
  (forall X, nds_tn_ax X n -> nds_tn_C X n <=c f)].
Proof.
move => nN.
rewrite /f /nds_tn_sup; set T := Zo _ _.
have TN: sub T Nat by apply: Zo_S.
have kN:=(NS_prod (NS_pow NS2 nN) (NS_factorial nN)).
set k := \2c ^c n *c factorial n.
have H: forall X, nds_tn_ax X n -> nds_tn_C X n <=c k.
  by move => X [ax _]; apply:(nds_tn_small nN) => i /ax /proj32_1.
have H1:forall X, nds_tn_ax X n -> natp (nds_tn_C X n).
  move => X x1; exact:(NS_le_nat (H _ x1) kN).
have ks:forall i, inc i T -> i <=c k by move => z /Zo_hi [X /H ax <-]. 
pose X (i:Set):= \1o.
have ax: nds_tn_ax X n.
  split; first by move => i _; apply: olt_01.
  exists \0o; [ apply: OS0 | move => i _; apply:odegree_one].
have ta: inc (nds_tn_C X n) T by  apply: (Zo_i (H1 _ ax)); exists X.
have ha: \0c <c union T.
  have: nonempty (nds_tn_S X n).
    set K:= n; have idp:=(permutation_id K).
    exists (nds_sc (cardinal K) (fun z => X (nth_elt K z)) (identity n)).
    apply/setUf_P; exists K; first by apply:setP_Ti.
    rewrite /nds_sc_Ktau/K (card_nat nN);apply/funI_P;ex_tac.
  move /card_nonempty1; rewrite -/(nds_tn_C X n) => h0.
  have cst:cardinal_set T by move => t /TN /CS_nat.
  exact:(clt_leT (card_ne0_pos  (cst _ ta) h0) (card_sup_ub cst ta)).
move: (Nat_sup_pr TN kN ks) => [pa pb pc pd];split => //; last first.
  move => Y x1; apply: pc; apply/ Zo_P; split; [by apply: H1 | by exists Y].
by case:pd => h1; [ empty_tac1 (nds_tn_C X n) | exact: (Zo_hi h1)].
Qed.
  
Lemma nds_tn_zero: nds_tn_sup \0c = \1c.
Proof.
move:(nds_tn_def NS0) => [_ _ [X ax <-] _].
rewrite /nds_tn_C /nds_tn_S setP_0 setUf_1 cardinal_set0 permutations_set0.
by rewrite funI_set1 cardinal_set1.
Qed.




Lemma nds_tn_prop2 n X:  natp n -> \0c <c n -> nds_tn_ax X n ->
  (nds_tn_S X n) =
  unionf n (fun a => fun_image (\Po (n -s1 a))
       (fun K => oopow  (odegree (X \0c)) *o (csumb (K +s1 a) (CNF_lc \o X)) 
           +o CNF_rem (X a)))
  +s1 \0c.
Proof.
move => nN np ax.
set ee :=  (odegree (X \0c)).
have nds_tn_axE i: i <c n ->
  [/\ \0o <o (CNF_lc (X i)), (CNF_lc (X i))  <o omega0,
    CNF_rem (X i) <o oopow ee &
    X i = oopow ee *o (CNF_lc (X i)) +o (CNF_rem (X i))].
   move => lin.
   move: ax => [ax1 [e ea eb]].
   by move: (the_CNF_split (ax1 i lin)); rewrite (eb _ lin) /ee (eb _ np).
have tn8 K s: 
  inc K (\Po n) -> inc s (permutations (cardinal K)) ->
  nonempty K ->
  nds_sc_Ktau X K s =
  oopow ee *o nds_sc_K (CNF_lc \o X) K
          +o CNF_rem (X (nth_elt K (Vf s \0c))).
  set m := cardinal K;move => /setP_P sb sp nek.
  move:(sub_smaller sb); rewrite (card_nat nN); rewrite -/m  => ckn.
  have mN:=(NS_le_nat ckn nN).
  have mz:= (card_ne0_pos (CS_nat mN) (card_nonempty1 nek)).
  pose c z := CNF_lc (X (nth_elt K z)).
  pose r z := CNF_rem (X (nth_elt K z)).
  have aux: forall i, i <c m ->
    [/\ \0o <o c i, c i <o omega0,  r i <o oopow ee
        & X (nth_elt K i) = oopow ee *o c i +o r i ]. 
    move => i l1.
    have sKN: sub K Nat by move => t /sb/(NltP nN) => h; apply(NS_lt_nat h nN).
    move: (nth_elt_inK (NS_lt_nat l1 mN) sKN l1) => /sb/(NltP nN).
    by move/nds_tn_axE.
  have h2: ordinalp ee.
    by move:(proj2 ax) => [e ea eb]; rewrite /ee (eb _ np).
  have h3 i: i <c m-> \0o <o c i /\ r i <o omega0 ^o ee by move =>  /aux []. 
  have h4 i: i <c m -> c i <o omega0  by move => /aux[].
  rewrite /nds_sc_Ktau /nds_sc_K -(nds_same_deg_s_perm mN mz h2 h3 h4 sp).
  apply:(osumf_exten mN)  => i /(NltP mN) lin.
  have /aux [_ _ _]//: (Vf s i) <c m.
  apply/(NltP mN).
  move/(permutationsP): sp => [[[fs _] _] ss ts]; Wtac.
have aux a K:inc K (\Po n) -> inc a K  ->
  nds_sc_K (CNF_lc \o X) K = csumb K (CNF_lc \o X). 
  move => pa pb; rewrite (nds_tn_of_nat nN _ pa) //.
  by move => i /nds_tn_axE [_ /(oltP OS_omega) h _ _].
have H s: nds_sc_Ktau X emptyset s = \0c.
    by rewrite  /nds_sc_Ktau cardinal_set0/nds_sc osum_f0. 
set_extens v.
  move/setUf_P => [K ka /funI_P [s sp sv]].
  case:(emptyset_dichot K) => ke; first by rewrite sv ke H; fprops.
  apply/setU1_P; left; apply /setUf_P.
  move: (nth_elt_prop7 nN ka sp ke); set a := (nth_elt K (Vf s \0c)) => ak.
  move/setP_P:(ka) => kb.
  have an: inc a n by apply: kb.
  ex_tac; apply /funI_P;exists (K -s1 a).
    by apply/setP_P => t/setC1_P [ha hb]; apply/setC1_P; split=> //; apply: kb.
  by rewrite sv (tn8 _ _ ka sp ke) (setC1_K ak) (aux _ _ ka ak).
case/setU1_P => h1; last first.
  rewrite h1; apply/setUf_P; exists emptyset; first by apply: setP_0i.
  apply/funI_P; exists (identity (cardinal emptyset)); last by rewrite H.
  apply:permutation_id. 
move/setUf_P: h1 => [a aN /funI_P[K /setP_P Kb ->]].
have nk1: ~  (inc a K) by move /Kb => /setC1_P[].
have eq1:= setU1_K nk1.
have ha: inc (K +s1 a) (\Po n).
   apply /setP_P => t/setU1_P; case; [by move/Kb => /setC1_P[] |  by move ->].
have hb:inc a (K +s1 a) by fprops.
have Kne: nonempty (K +s1 a) by ex_tac.
move: (nth_elt_prop8 nN ha hb) => [f fp aaa].
apply/setUf_P; ex_tac;apply/funI_P; ex_tac.
by rewrite (tn8 _ _ ha fp Kne) - aaa (aux _ _ ha hb).
Qed.


Lemma nds_tn_max X n: natp n -> \0c <c n -> nds_tn_ax X n ->
  (nds_tn_C X n) <=c ndsC n.
Proof.
move => nN np ax.
rewrite /ndsC /nds_tn_C (nds_tn_prop2 nN np ax); set F:= (fun a : Set => _).
suff h: cardinal (unionf n F) <=c  (n *c \2c ^c cpred n).
  have w:cardinalp (n *c \2c ^c cpred n) by fprops.
  case: (inc_or_not \0c (unionf n F)) => ha. 
    rewrite (setU1_eq ha); apply:(cleT h (cleS0 w)).
  by rewrite (csucc_pr ha); apply /(cleSSP (CS_cardinal _) w).
apply: (cleT(csum_pr1_bis n F));rewrite cprodC - csum_of_same.
apply:csum_increasing; bw => a abn; bw.
rewrite - (nds_C_prop nN abn); apply:fun_image_smaller.
Qed.

Definition sumpow2 K := csumb K (fun z => \2c ^c z).

Lemma sumpow2_N1 K: finite_set K -> sub K Nat -> natp (sumpow2 K).
Proof.
move => ha hb; apply:finite_sum_finite; split; aw; bw.
hnf; bw => x xk; bw; apply:(NS_pow NS2 (hb _ xk)).
Qed.

Lemma sumpow2_N2 n a K:
  natp n -> inc K (\Po (n -s1 a)) -> natp (sumpow2 K).
Proof.
move => nN /setP_P h1; move:(sub_trans h1 (@sub_setC _ _)) => h2.
have h3:= (sub_trans h2 (NS_inc_nat nN)).
exact (sumpow2_N1 (sub_finite_set h2 (finite_set_nat nN)) h3).
Qed.

Lemma sumpow2_rec a K: inc a K -> 
  sumpow2 K = sumpow2 (K -s1 a) +c (\2c ^c a). 
Proof. 
move => aK; rewrite -{1} (setC1_K aK) /sumpow2 - (csumA_setU1)//.
by case/setC1_P. 
Qed.

Lemma sumpow2_bd n K: 
   natp n -> (forall i, inc i K -> i <c n) ->
   sumpow2 K <c \2c ^c n.
Proof.
move => nN;move:n nN K ; apply: Nat_induction.
  move => K h; rewrite cpowx0.
  case:(emptyset_dichot K). 
     move => ->; rewrite /sumpow2/ csumb csum_trivial;bw;apply:clt_01.
  by move => [i /h /clt0].
move => n nN Hrec K kP.
set K1:= K -s1 n.
have /Hrec r1:(forall i, inc i K1 -> i <c n). 
   by move => i /setC1_P [/kP /(cltSleP nN) h1 h2].
rewrite (cpow_succ _ nN) cprodC two_times_n.
case: (inc_or_not n K) => ink.
  rewrite (sumpow2_rec ink); apply:csum_Mlteq; fprops.
rewrite -(setC1_eq ink); apply: (clt_leT r1); apply:csum_M0le; fprops.
Qed.

Lemma sumpow2_inj n K1 K2 : natp n ->
  (forall i, inc i K1 -> i<c n) -> (forall i, inc i K2 -> i<c n) ->
   sumpow2 K1 = sumpow2 K2 -> K1 = K2.
Proof.
move => nN; move: n nN K1 K2; apply: Nat_induction.
  move => K1 K2 h1 h2 sv.
  case:(emptyset_dichot K1) => ha.
      case:(emptyset_dichot K2) => hb; first by rewrite ha hb.
     by move:hb => [i/h2 /clt0].
   by move:ha => [i/h1 /clt0].
move => n nN Hrec K1 K2 h1 h2.
have kk1:(forall i, inc i (K1 -s1 n) -> i <c n). 
  by move => i /setC1_P [/h1 /(cltSleP nN) hu hv].
have kk2:(forall i, inc i (K2 -s1 n) -> i <c n). 
  by move => i /setC1_P [/h2 /(cltSleP nN) hu hv].
have l1:(sumpow2 (K1 -s1 n)) <c \2c ^c n by apply: sumpow2_bd.
have l2:(sumpow2 (K2 -s1 n)) <c \2c ^c n by apply: sumpow2_bd.
have pN:= (NS_pow NS2 nN).
case: (inc_or_not n K1) => nk1.
   case: (inc_or_not n K2) => nk2.
      have hh1:= (NS_lt_nat l1 (NS_pow NS2 nN)).
      have hh2:= (NS_lt_nat l2 (NS_pow NS2 nN)).
      rewrite (sumpow2_rec nk1) (sumpow2_rec nk2) => eq2.
      move:(csum_eq2r pN hh1 hh2 eq2) => eq3.
      by rewrite - (setC1_K nk1) - (setC1_K nk2) (Hrec _ _ kk1 kk2 eq3).
  move: l2; rewrite (sumpow2_rec nk1) (setC1_eq nk2) => l2 eq.
  move: (csum_M0le (sumpow2 (K1 -s1 n)) (CS_nat pN)).
  by rewrite csumC eq => /(cltNge l2).
case: (inc_or_not n K2) => nk2.
  move: l1; rewrite (setC1_eq nk1)  (sumpow2_rec nk2) => l1 eq.
  move: (csum_M0le (sumpow2 (K2 -s1 n)) (CS_nat pN)).
  by rewrite csumC - eq =>  /(cltNge l1).
by rewrite - (setC1_eq nk1) - (setC1_eq nk2); apply: Hrec.
Qed.


Lemma omega_monom_inj a b c d: 
  natp a -> natp b -> natp c -> natp d ->
  omega0 *o a +o b = omega0 *o c +o d -> (a = c /\ b = d).
Proof.  
move => aN bN cN dN eq1.
have aux: forall u v, natp u -> natp v -> natp (omega0 *o u +o v) -> u = \0c. 
  move => u v uN vN /(oltP OS_omega) h; ex_middle h1.
  have l1:= (oprod_Mle1 OS_omega (ord_ne0_pos (OS_nat uN) h1)).
  by case:(oleNgt (oleT l1 (osum_Mle0 (proj32 l1) (OS_nat vN)))).
case: (czero_dichot (CS_nat aN))=> az.
  move: eq1; rewrite az oprod0r (osum0l (OS_nat bN)) => eq1.
  rewrite eq1 in bN.
  by rewrite eq1 (aux _ _ cN dN bN)  oprod0r (osum0l (OS_nat dN)). 
case: (czero_dichot (CS_nat cN))=> cz.
  move: eq1; rewrite cz oprod0r (osum0l (OS_nat dN)) => eq1.
  case: (nesym (proj2 az));rewrite - eq1 in dN; exact:(aux _ _ aN bN dN). 
move:(the_CNF_omega_kj aN bN az) => [_ [_ e1 e2]]. 
move:(the_CNF_omega_kj cN dN cz) => [_ [_ e3 e4]]. 
by rewrite eq1 in e1 e2; rewrite e1 in e3; rewrite  e2 in e4.
Qed.


Definition nds_base := fun i => omega0 *o (\2c ^c i) +o i.

Lemma nds_base_prop i (u := nds_base i) : natp i ->
  [/\ odegree u = \1c, CNF_lc u = (\2c ^c i) & CNF_rem u = i].
Proof.
move => iN.
by case:  (proj2 (the_CNF_omega_kj (NS_pow NS2 iN) iN (cpow2_pos i))).
Qed.


Lemma nds_tn_ex_ax n:  natp n -> nds_tn_ax nds_base n.
Proof.
move => nN; split.
  move => i lin; move:(NS_lt_nat lin nN) => yN.
  exact:(proj1 (the_CNF_omega_kj (NS_pow NS2 yN) yN (cpow2_pos _))). 
exists \1o; first by apply:OS1.
by move => i lin; move:(NS_lt_nat lin nN) => /nds_base_prop [].
Qed.


Lemma nds_tn_ex_small n:
  natp n -> nds_rel_small_ax \1c (\2c ^c (cpred n)) nds_base n. 
Proof.
have p2p k: \0o <o \2c ^c k by  exact:(oclt(cpow2_pos k)).
rewrite/nds_base => nN k lkn.
have kN:= NS_lt_nat lkn nN.
case: (ozero_dichot (OS_nat kN)) => kz.
  rewrite kz cpowx0 (oprod1r (OS_omega)) (osum0r OS_omega).
  move:(rel_small_omega); apply:rel_small_le2; apply/oge1P; exact: p2p.
have clo: \2c ^c k <o omega0 by apply/olt_omegaP; fprops.
have lko: k <o omega0 by  apply/olt_omegaP.
move: (@CNF_two_terms \1o (\2c ^c k ) \0o k olt_01 (p2p k) clo kz lko).
rewrite /rel_small_ordinal /odegree.
rewrite oopow1 oopow0 (oprod1l (proj31_1 lko)); set u:=  the_CNF _.
move => [axu uv [-> e0 e1 c0 c1]].
rewrite cpred2 e1; split; first by apply:(oleR OS1).
have le1: \2c ^c k <=c \2c ^c cpred n.
  move: (cpred_pr nN (nesym (proj2(cle_ltT (cle0n kN) lkn)))) => [pa pb].
  by apply: (cpow_Meqle (card2_nz)); apply/(cltSleP pa); rewrite - pb.
have  le2:= (cleT (proj1 (cantor (CS_nat kN)))  le1).
by move => j /clt2; case; move ->;rewrite ? c0 ? c1 => //; apply: ocle.
Qed.

(* deplacer *)
Lemma funI_exten f1 f2 X:
   {inc X, f1 =1 f2} -> fun_image X f1 = fun_image X f2.
Proof.
by move => hyp; set_extens t ; move => /funI_P [y ya ->];
   apply/funI_P; ex_tac; symmetry;apply: hyp.
Qed.


Lemma nds_tnmax_ex n: natp n -> 
  (nds_tn_S nds_base n) =
  unionf n (fun a => fun_image (\Po (n -s1 a))
    (fun K => omega0 *o (sumpow2 (K +s1 a)) +o a)) +s1 \0c.
Proof. 
move => nN.
have npr:=(NS_inc_nat nN).
case: (czero_dichot (CS_nat nN)) => np.
  rewrite /nds_tn_S np setP_0 setUf_0 set0_U2 setUf_1 cardinal_set0.
  apply: set1_pr1.
    apply: funI_setne; move: (permutation_id \0c) => h; ex_tac.
  move => i /funI_P [z zz ->].
  by rewrite /nds_sc_Ktau /nds_sc cardinal_set0 osum_f0.
rewrite (nds_tn_prop2 nN np (nds_tn_ex_ax nN)).
f_equal; apply: setUf_exten => a /npr an; apply: funI_exten => K KP.
rewrite (proj33 (nds_base_prop an)) (proj31 (nds_base_prop NS0)).
rewrite oopow1; apply: f_equal2; [apply: f_equal | exact].
apply: csumb_exten => t tK; apply: (proj32 (nds_base_prop _)).
move/setP_P: KP => KP.
by case/setU1_P: tK; [ move/KP => /setC1_P [/npr] |  move->].
Qed.


Lemma nds_tn_ex_val n: natp n ->  nds_tn_C nds_base n = ndsC n.
Proof.
move => nN.
have ha: forall i, natp i -> natp (\2c ^c i).
  move => i; apply:(NS_pow NS2).
rewrite /nds_tn_C (nds_tnmax_ex nN).
have sk2_S a K: inc K (\Po (n -s1 a)) -> 
   sumpow2 (K +s1 a) = sumpow2 K +c \2c ^c a. 
  move => /setP_P KS.
  by rewrite /sumpow2 (csumA_setU1) // => /KS /setC1_P [].
have sk2_N a K: inc a n ->  inc K (\Po (n -s1 a)) -> natp (sumpow2 (K +s1 a)).
  move => aN KP. rewrite (sk2_S a K KP); apply:(NS_sum (sumpow2_N2 nN  KP)).
  exact:(NS_pow NS2 (NS_inc_nat nN aN)).
rewrite csucc_pr; last first.
  move/setUf_P => [a an /funI_P [K Ka]]; apply/nesym.
  have aN :=  (NS_inc_nat nN an).
  move:(OS_nat (sk2_N _ _ an Ka)) => o1.
  move/(osum2_zero (OS_prod2 OS_omega o1) (OS_nat aN)) => /proj1.
  apply:(oprod2_nz OS_omega o1 omega_nz); rewrite (sk2_S _ _ Ka) => /csum_nz.
  by move => [_] /cpow2_nz.
rewrite /ndsC; congr csucc;rewrite csum_pr4_bis; last first.
  move => i j iN jN; case (equal_or_not i j) => eij; [by left | right].
  apply:disjoint_pr => u /funI_P [K1 K1p uv1] /funI_P [K2 K2p uv2]; case eij.
  rewrite uv1 in uv2.
  move: (sk2_N _ _ iN K1p)  (sk2_N _ _ jN K2p) => e2 e4.
  by case: (omega_monom_inj e2  (NS_inc_nat nN iN) e4 (NS_inc_nat nN jN) uv2).
set F := fun a:Set => _.
suff H: forall a, inc a n -> F a = \2c ^c cpred n.
   rewrite - {2} (card_nat nN)  (cprod2cl). 
   by rewrite cprodC - csum_of_same; apply: csumb_exten => i ib /=; apply: H.
move => a an; rewrite - (nds_C_prop nN an) /F cardinal_fun_image //.
move => K1 K2 K1a K2a /= eq1.
move:(sk2_N _ _ an K1a) (sk2_N _ _ an K2a) => s1N s2N.
have aN := (NS_inc_nat nN an).
move: (proj1 (omega_monom_inj s1N aN s2N aN eq1)).
rewrite (sk2_S _ _ K1a) (sk2_S _ _ K2a) => eq2.
move:(csum_eq2r (NS_pow NS2 aN) (sumpow2_N2 nN K1a) (sumpow2_N2 nN K2a)  eq2).
apply: (sumpow2_inj nN) => i.
  by move/setP_P: K1a => h; move=> /h /setC1_P /proj1/(NltP nN).
  by move/setP_P: K2a => h; move=> /h /setC1_P /proj1/(NltP nN).
Qed.


Lemma nds_tn_value n: natp n -> nds_tn_sup n = ndsC n.
Proof.
move => nN.
case: (czero_dichot (CS_nat nN)) => np.
  by rewrite np nds_tn_zero /ndsC cprodC cprod0r succ_zero.
move: (nds_tn_def nN) => [_ _ [X hx eq1] hu].
apply: cleA; first by rewrite - eq1; exact:(nds_tn_max nN np hx).
rewrite - (nds_tn_ex_val nN); apply: (hu _ (nds_tn_ex_ax nN)).
Qed.

Lemma osumf_mul c n X: natp n -> nds_ax X n -> ordinalp c ->
   osumf (fun z => c *o X z) n = c *o osumf X n.
Proof.
move:n; apply: Nat_induction.
  by rewrite !osum_f0 (oprod0r).
move => n nN Hrec axn oc.
move:(true_below_rec nN axn) => [ax1 oXn].
by rewrite !(osum_fS _ nN) (Hrec ax1 oc) (oprodD oXn (OS_osumf nN ax1) oc).
Qed.


Lemma nds_card_mul c n X (Y:= fun i => c *o X i):
  natp n -> nds_ax X n -> \0o <o c ->
  nds_card n X = nds_card n Y.
Proof.
move => nN ax cp.
rewrite /nds_card /nds_sums.
set E := permutations n; set A := fun_image _ _;set B := fun_image _ _.
apply /card_eqP; exists (Lf (fun z => c *o z) A B).
have aux: forall f, inc f E ->  nds_sc n Y f = c *o nds_sc n X f.
  move => f /(nds_ax_perm nN ax) [ax1 _ ]. 
  apply:(osumf_mul nN ax1 (proj32_1 cp)).  
have osA: ordinal_set A.
  by move => z /funI_P [f /(nds_ax_perm nN ax) [_ os] ->].
split; aw.
apply: lf_bijective.
+ by move => z /funI_P [f fp ->]; apply /funI_P; exists f => //; rewrite aux.
+ move => u v /osA ou /osA ov h; exact (oprod2_simpl ou ov cp h).
+ move => z /funI_P [f fp ->]; exists(nds_sc n X f). 
  apply /funI_P; exists f => //.
  by apply: aux.
Qed.

Lemma nds_small_mul d c e n X (Y:= fun i => (oopow e) *o X i):
  natp n -> nds_ax X n -> ordinalp e -> ordinalp c ->
  nds_rel_small_ax d c X n -> nds_rel_small_ax ( e +o d) c Y n.
Proof.
move => nN ax oe oc sm i lin; move: (sm i lin)(ax i lin) => [dx cxs] oxi.
rewrite /Y.
have od := proj32 dx.
case: (ozero_dichot oxi) => xiz.
  rewrite xiz oprod0r; apply: (rel_small_zero (OS_sum2 oe od) oc).
split.
  move: (odegree_prod (oopow_pos oe) xiz); rewrite (odegree_opow oe) => ->.
  apply: (osum_Meqle dx oe).
move: (the_CNF_p0 oxi);set X0 := (the_CNF (X i)).
rewrite /CNFB_ax.
move => [[ _ lN _ [_ _ _ _ [_ _ cax]]] bv].
move: (CNF_prod_pr0 lN cax oe); rewrite -/(CNFBv _) bv;move => [HA HB].
move: (the_CNF_p3 HA lN); rewrite HB => ->.
rewrite /CNFbB !pr2_pair !pr1_pair => j ljd.
bw; [ by Ytac0; apply: cxs | apply: (NS_lt_nat ljd lN)].
Qed.

  
Definition nds_type_nor X n k e:=
 [/\ nds_type X n k, ordinalp e,
    forall i, i<c n -> \0o <o (X i),
    forall i, i<c k -> odegree (X i) = e  &
    forall i, k <=c i -> i <c n -> e <o odegree (X i)].

(* dire y petit *)


Lemma nds_type_nor_ex X n k: 
   natp n -> \0c <c k -> k <=c n -> nds_type X n k ->
   exists Y e,  [/\ nds_same_dc_bound n X Y, nds_type_nor Y n k e &
                      nds_card n Y = nds_card n X].
Proof.
move => nN knz lekn [ha [m [om h1]]].
set E := Zo _ _ => ce.
have ax:nds_ax X n by move => i /ha/proj32_1.
have sk: sub  k n by case: lekn.
have kN:=(NS_le_nat lekn nN).
have sE: sub E n by apply: Zo_S.
move:(permutation_exists2 nN sE) => [f fp]; rewrite ce.
move: (nds_card_perm_inv nN ax fp); set Y:= fun i =>X (Vf f i) => yv fv.
rewrite yv.
exists Y,m; split; last reflexivity.
  exact: (nds_same_dc_bound_perm X nN fp).
move:fp => /Zo_P [/functionsP [ff sf tf] bf] .
have sisf: sub  k (source f) by rewrite sf.
move: (cardinal_image sisf (proj1 bf));rewrite- fv (card_nat  kN) => ck.
have pa i: i <c n -> \0o <o Y i.
  move => /(NltP nN) ii; apply: ha; apply/(NltP nN); Wtac.
have pb i: i <c n -> m <=o odegree (Y i).
  move => /(NltP nN) ii; apply: h1;apply/(NltP nN); Wtac.
have pc:  forall i, i <c k -> odegree (Y i) = m.
  move => i /(NltP kN) ik. 
  have/Zo_hi //: inc (Vf f i) E
    by rewrite fv; apply/(Vf_image_P ff sisf); ex_tac.
have pd: (Zo n (fun i => odegree (Y i) = m)) = k.
  set_extens t. 
    move => /Zo_P [ta tb].
    have: inc (Vf f t) E by apply:Zo_i; [Wtac | apply: tb ].
    rewrite fv => /(Vf_image_P ff sisf) [u usf].
    move: ta (sk _ usf); rewrite - sf => hb hc hd.
    by rewrite(proj2(proj1 bf) _ _ hb hc hd).
  move => ta; apply /Zo_P;split; first by apply: sk.
  by apply: pc; apply /(NltP kN). 
have pe i: k <=c i -> i <c n -> m <o odegree (Y i).
  move => ki kin; split;[ by apply: pb | move => sa ].
  have: inc i k.
    by rewrite - pd; apply /Zo_P; split; [by apply/(NltP nN) | done].
   by move/(NltP kN) => /(cleNgt ki).
have pf:nds_type Y n k.
  split;[ exact|  exists m]; rewrite pd (card_nat kN); split; exact.
split; exact.
Qed.


Lemma nds_tn_C_exten Z Y n: natp n -> same_below Z Y n ->
  nds_tn_C Z n = nds_tn_C Y n.
Proof.
move => nN h.
have aux: forall K s, inc K (\Po n) ->
    inc s (permutations (cardinal K)) -> nds_sc_Ktau Z K s = nds_sc_Ktau Y K s.
  move => K s /setP_P Kb /permutationsP [[[fs _] _ ] ss tf].
  have fsK:=(sub_finite_set Kb (finite_set_nat nN)).
  have ckN: natp (cardinal K) by apply /NatP.
  have sKN:=(sub_trans Kb (NS_inc_nat nN)).
  apply:(osumf_exten ckN) => i lin; apply: h.
  have h1: Vf s i <c cardinal K.
     by apply /(NltP ckN); Wtac; rewrite ss;apply /(NltP ckN).
  apply/(NltP nN); apply: Kb; apply: (nth_elt_inK (NS_lt_nat h1 ckN) sKN h1).
rewrite /nds_tn_C /nds_tn_S; apply: f_equal.
set_extens t; move => /setUf_P [K KP /funI_P [s sp ->]]; 
  apply /setUf_P; ex_tac;apply/funI_P; ex_tac. 
  symmetry;exact:(aux K s KP sp).
Qed.
  
Section NdsStudy.
Variables (X: fterm) (n k e: Set).
Hypothesis nN: natp n.
Hypothesis knz: \0c <c k. 
Hypothesis kln: k <c n.
Hypothesis Xax: nds_type_nor X n k e.

Let pL f i := k <=c Vf f i.
Let pS f i := Vf f i <c k.
Let fL f := intersection (Zo Nat (fun i => i <c n /\ pL f i)).

Lemma nds_fL_prop f: inc f (permutations n) -> 
  [/\ natp (fL f), (fL f) <=c k,   (fL f) <c n,  pL f (fL f) & 
      forall j, j <c (fL f) -> pS f j].
Proof.
move => fp; rewrite /fL; set E:= Zo _ _.
move/permutationsP:(fp) =>[[injf [ff fs]] sf tf].
have nee: nonempty E.  
  move:(perm_int_surj nN fp kln) => [i lin si].
  exists i;apply:Zo_i;[ by apply: (NS_lt_nat lin nN) | split => //].
  rewrite /pL si; apply: (cleR (proj31_1 kln)).
have seN: sub E Nat by apply:Zo_S.
move: (inf_Nat seN nee); set i:= intersection E. 
move => [/Zo_P [iN [ha hb]] etc].
have kN:= (NS_lt_nat kln nN).
have aux:  forall j, j <c i -> pS f j.
  move => j lji.
  move: (clt_ltT lji ha) => ljn; move/(NltP nN): (ljn) => jb.
  have /(NltP  nN) /proj31_1 cf: (inc (Vf f j) n) by Wtac.
  case:   (cleT_el (proj31_1 kln) cf) => // jfL; case:(cltNge lji).
  by apply: etc; apply:(Zo_i  (NS_lt_nat ljn nN)).
split => //.
case: (NleT_el iN kN) => // lki.
have aux2: forall i, i <=c k -> pS f i. 
  by move =>j jk; apply: aux; exact (cle_ltT jk lki).
rewrite /pS in aux2.
have h3: sub (Nintc k) (source f).
  move => j/(NintcP kN) ja; rewrite sf; apply/(NltP nN);exact (cle_ltT ja kln).
have su1: sub (Vfs f (Nintc k)) k.
  move => t /(Vf_image_P ff h3) [u /(NintcP kN) ul1 ->].
  by apply/(NltP kN); apply: aux2.
move: (sub_smaller su1).
rewrite(cardinal_image h3 injf) (card_Nintc kN) (card_nat kN) => l1.
case: (cleNgt l1 (cltS kN)).
Qed.

Lemma nds_fL_prop2 f m (q:= (fL f) +c m) :
  inc f (permutations n) -> natp m -> q <c n ->
  (forall j, j<c m -> pL f ((fL f) +c j)) ->
  [\/ (m = n -c k /\ forall j, j <c n -> q <=c j -> pS f j),
     pL f q |
    q <> \0c /\
    exists p, [/\ natp p, csucc (q +c p) <c n,
       pL f (csucc (q +c p)),
       pL f (cpred q) &
       forall z, z <=c p -> pS f (q +c z) ]].
Proof.
move => fp mN inb Hi.
move: (nds_fL_prop fp) => [sa _ sb sc sd].
move/permutationsP: fp => [bf sf tf].
have ff: function f by fct_tac.
have qN: natp q by apply: NS_sum.
have kN:= (NS_lt_nat kln nN).
set E1 := Zo Nat (fun z => [/\ z <c n, pL f z & q <=c z]).
set E2:= Zo Nat (fun z => [/\ z <c n, pL f z & z <c q]).
set E3:= n -s k.
have: disjoint E1 E2. 
  by apply: disjoint_pr => u /Zo_P[_ [_ _ ha]]/Zo_P[_ [_ _ /(cleNgt ha)]].
move/csum2_pr5;rewrite - csum2cl - csum2cr.
have E3tf: sub E3 (target f) by rewrite tf;move => t /setC_P [].
have ->: E1 \cup E2 = fun_image E3 (Vf (inverse_fun f)).
   set_extens t => h.
    have[ha hb]: t <c n /\ k <=c Vf f t by case/setU2_P:h => /Zo_P[_ [ha hb _]].
    have tsf: inc t (source f) by rewrite sf; apply /(NltP nN).
    have se:=(inverse_V2 bf tsf).
    apply/funI_P; exists (Vf f t); last by exact.
    apply /setC_P; split; [ Wtac | by move/(NltP kN) => /(cleNgt hb)].
  move/funI_P:h => [z zi ->]; move: (E3tf _ zi) => ft.
  have eq2:=(inverse_V bf ft).
  have zt:= (inverse_Vis bf  ft).
  rewrite sf in zt; move /(NltP nN): zt => lt1.
  have izN:= (NS_lt_nat lt1 nN).
  move/setC_P: zi => [/(NS_inc_nat nN) zN hb].
  case: (NleT_el kN zN) => zz; last by case: hb; apply /(NltP kN).
  by case: (NleT_el qN izN) => lea; apply /setU2_P; [left | right];
     apply:Zo_i => //; split => //; rewrite /pL eq2.
have ->: E2 = fun_image m (fun z => (fL f) +c  z).
  set_extens t.
    move/Zo_P => [tN [tn h tq]]; apply/funI_P.
    case: (cleT_el (proj31_1 sb)(proj31_1 tq)) => le1.
      move:tq; rewrite - (cdiff_pr le1) => tq.
      move /(NltP mN):(csum_lt2l sa (NS_diff (fL f) tN) mN tq).
      move => di; ex_tac.
    by move: h (sd _ le1);rewrite /pL /pS => ua /(cleNgt ua).
  move /funI_P => [z /(NltP mN) zi ->]; apply/Zo_P.
  move: (clt_ltT (csum_Meqlt sa zi) inb) => lt1.
  move: (csum_Meqlt sa zi) => sf1.
  split; [ exact:(NS_lt_nat lt1 nN) | by split => //; apply: Hi ].
rewrite cardinal_fun_image; last first.
  by move => i j /E3tf /(inverse_V bf) {2}<- /E3tf  /(inverse_V bf) {2}<- ->.
rewrite cardinal_fun_image; last first.
  by move => i j im jm /=; apply:(csum_eq2l sa); apply: (NS_inc_nat mN).
rewrite /E3 -/(_ -c _) (card_card (CS_nat mN)) => cc.
have cE1N: natp (cardinal E1).
  by  move:(NS_diff k nN); rewrite cc => /(Nat_in_suml (CS_cardinal E1)).
have ci:= (CS_nat mN).
case: (cleT_ell (CS_diff n k) ci) => cink.
+ constructor 1; split; first by symmetry.
  move => j ljn lqj.
  have jsf: inc j (source f) by rewrite sf; apply /(NltP nN).
  move:(Vf_target ff jsf); rewrite tf; move/(NltP nN) => ha.
  case: (cleT_el (proj31_1 kln) (proj31_1 ha)) => hb; last exact.
  rewrite cink - {1} (csum0l ci) in cc.
  move:(esym(csum_eq2r mN NS0 cE1N cc))=>/card_nonempty hw.
  have jE1: inc j E1 by apply:Zo_i; [apply: (NS_lt_nat ljn nN) | ].
  empty_tac1 j.
+ by move: (csum_M0le (cardinal E1) ci);rewrite csumC - cc=> /(cltNge cink).
+ have lt1: Vf f q <c n by apply/(NltP nN); Wtac; rewrite sf;apply/(NltP nN).
  case: (cleT_el (CS_nat kN) (proj31_1 lt1)) => lt2; first by constructor 2.
  constructor 3.
  case: (equal_or_not m \0c) => inz.
    by move:sc lt2; rewrite /pL /q inz (csum0r (proj31_1 sb))=> ua /(cleNgt ua).
  split; first by move/csum_nz; case. 
  case: (emptyset_dichot E1) => nee.
    by move: (proj2 cink); rewrite cc nee cardinal_set0 (csum0l ci).
  have seB: sub E1 Nat by apply:Zo_S.
  move: (inf_Nat seB nee) =>[]; set j:= intersection E1 => ja jb.
  move/Zo_P: ja => [jN [ljn jc lqj]].
  have eq0:= (cdiff_pr lqj).
  case: (equal_or_not (j -c q) \0c) => eq1.
    by move: jc; rewrite - eq0 eq1 (csum0r (proj31 lqj)) /pL => /(cltNge lt2).
  move: (cpred_pr (NS_diff _ jN) eq1)=> []. 
  set p :=  (cpred (j -c q)) => pN pv.
  have eq4:q +c csucc p = j by rewrite - pv.
  have eq2: csucc (q +c p) = j by rewrite - (csum_nS _ pN).
  move: (cpred_pr mN inz) => [ia ib].
  have eq3: q = csucc (fL f +c (cpred m)) by rewrite -(csum_nS _ ia) - ib.
  move: (cpred_pr1 (CS_sum2 (fL f) (cpred m))); rewrite - eq3 => eq5.
  move: (cltS ia); rewrite - ib => /Hi; rewrite - eq5 => eq6.
  exists p; rewrite eq2; split => //.
  move => z /(cltSleP pN) zm.
  move:(csum_Meqlt qN zm); rewrite eq4 => l1.
  move: (clt_ltT l1 ljn) => ra.
  move: (NS_lt_nat ra nN) => qzN.
  move/(NltP nN):(ra); rewrite - sf => /(Vf_target ff); rewrite tf.
  move /(NltP nN) => /proj31_1 fzn.
  case: (cleT_el (CS_nat kN) fzn) => // l2.
  have: (inc (q +c z) E1). 
     apply:Zo_i => //; split => //; apply:(Nsum_M0le _ qN).
  by move/jb => /(cltNge l1).
Qed.

Lemma nds_fL_prop3 i l f:
  inc f (permutations n) -> natp i -> natp l ->
  (csucc (csucc (i +c l))) <c n ->
  pL f i -> 
  (forall j, j <=c l -> pS f (csucc (i+c j))) ->
  pL f (csucc (csucc (i +c l))) ->
  exists g, 
     [/\ inc g (permutations n), nds_sc n X f = nds_sc n X g,
        forall j, j<c (csucc i) -> Vf f j = Vf g j & pL g (csucc i)].
Proof.
move => fp iN lN ltil2n p1 p2 p3.
set i1:= csucc i; set il2:= csucc (csucc (i +c l)); set il3:= csucc il2.
set il:= csucc (i +c l).
rewrite -/il2 in ltil2n.
have il1N:= NS_succ (NS_sum iN lN).
have il2N:= (NS_lt_nat ltil2n nN).
have il3N: inc il3 Nat by apply:NS_succ il2N. 
have i1N:= NS_succ iN.
have i2N:= NS_succ i1N.
have leil3n: il3 <=c n by  apply/cleSltP.
have lti1il2: i1 <c il2.
  rewrite /i1 /il2; apply:cltSS. 
  exact (cle_ltT (Nsum_M0le l iN) (cltS (NS_sum iN lN))). 
have lti1n:=(clt_ltT lti1il2 ltil2n).
have ltin:= (clt_ltT (cltS iN) lti1n).
have Ha: inc il2 n by apply /(NltP nN).
have Hb: inc i1  n by apply /(NltP nN).
have le1:=(cleS (NS_succ iN)).
move: (transposition_prop nN Ha Hb).
set g := Lf _ _ _.
move => [gp gssi gsi ngsi ggi].
move/permutationsP: (gp) => [[[fg _] _] sg tg].
move/permutationsP: (fp) => [[[ff _] _] sf tf].
set h:= f \co g.
have fgP: f \coP g by split => //; ue.
have pa: forall j, j <c csucc i -> Vf f j = Vf h j.
  move => j jsi.
  have jb: inc j n by apply /(NltP nN); exact : (clt_ltT jsi lti1n).
  have jb': inc j (source g) by ue.
  have ne1:= (proj2(clt_ltT jsi (lti1il2))).
  by rewrite /h; aw; rewrite ngsi //; exact (proj2 jsi).
have pb: forall j, inc j  n -> il3 <=c j -> Vf f j = Vf h j.
  move => j jsi l3.
  have jb': inc j (source g) by ue.
  have lt1:= (clt_leT (cltS il2N) l3).
  have ne2:= (nesym (proj2 (clt_ltT lti1il2 lt1))).
  have ne1:= nesym(proj2 lt1).
  by rewrite /h; aw;rewrite ngsi.
have hp:= (permutation_Sc fp gp).
have hh: pL h i1 by rewrite /h /pL; aw; [ rewrite gsi | rewrite sg].
suff: nds_sc n X f = nds_sc n X h by exists h; split; exact.
move:  Xax => [_ _ hw hu hv].
have ax: nds_ax X n by move => z /hw /(proj32_1).
move: (nds_ax_perm nN ax fp) (nds_ax_perm nN ax hp).
rewrite /nds_sc. 
set Y:= (fun z => X (Vf f z)).
set Z:= (fun z => X (Vf h z)); move => [aY _ ] [aZ _].
move: (cdiff_pr leil3n); set m := _ -c _ => mv.
have mN: inc m Nat by apply: (NS_diff _ nN).
rewrite - mv in aY aZ |- *. 
rewrite (osum_fA il3N mN aY) (osum_fA il3N mN aZ).
apply: f_equal2.
   apply: (osumf_exten mN) => z zm.
   have le2:il3 <=c z +c il3 by apply: csum_Mle0; fprops.
   have le3: inc (z +c il3) n. 
     by apply/(NltP nN); rewrite - mv (csumC il3); apply:csum_Mlteq.
   by rewrite /Y/Z pb.
set a := X (Vf f i); set b := X(Vf f i1); set c:= X(Vf f il2).
have il2sg: inc il2 (source g) by  ue.
have i1sg: inc i1 (source g) by  ue.
have ib: inc i n by apply /(NltP nN).
have ig: inc i (source g) by ue.
rewrite 2! (osum_fS _ il2N).
rewrite {1} /Z {1}/Y /h (compf_V fgP il2sg) gssi -/b -/c.
have le2: csucc (csucc i) <=c il2.
  apply/(cleSSP (CS_succ i) (CS_succ ((i +c l)))).
  apply/(cleSSP (CS_nat iN) (CS_sum2 i l)).
  apply: (Nsum_M0le _ iN).
move: (cdiff_pr le2); set m1 := _ -c _ => m1v.
have m1N: inc m1 Nat by apply: (NS_diff _ il2N).
rewrite mv in aY aZ.
have aY': nds_ax Y (csucc (csucc i) +c m1).
  move => z; rewrite m1v => l1; apply: aY; exact:(clt_ltT l1 ltil2n).
have aZ': nds_ax Z (csucc (csucc i) +c m1).
  move => z; rewrite m1v => l1; apply: aZ; exact:(clt_ltT l1 ltil2n).
have ha:= cltS iN.
have hb:= (proj2 (clt_ltT ha lti1il2)).
rewrite - m1v (osum_fA i2N m1N aY') (osum_fA i2N m1N aZ').
rewrite 2! (osum_fS _ i1N) 2! (osum_fS _ iN).
rewrite {2 3} /Y {2 3} /Z /h (compf_V fgP ig)(compf_V fgP i1sg).
rewrite gsi (ngsi _ ib hb (proj2 ha)) -/a -/b -/c.
set A:= osumf _ m1.
set B:= osumf _ i.
have <-: A = (osumf (fun z : Set => Z (z +c csucc (csucc i))) m1).
  apply: (osumf_exten m1N) => j lej.
  have [_ /nesym ltb]:i1 <c j +c csucc (csucc i).
     apply/(cleSltP i1N); apply:csum_Mle0; fprops.
  have lta: j +c csucc (csucc i) <c il2.
     rewrite - m1v (csumC _ m1); exact: (csum_Mlteq i2N lej).
  have hc: inc (j +c csucc (csucc i)) (source g).
    rewrite sg; apply/(NltP nN); rewrite - mv /il3 - m1v.
    rewrite (csum_Sn _ (NS_sum i2N m1N)) (csumC _ m1).
    apply: (clt_ltT lta); rewrite (csumC m1) m1v.
    apply /(cltSleP (NS_sum il2N mN)) /csum_M0le; fprops.
  by rewrite /Y/Z /h (compf_V fgP hc) (ngsi _ _ (proj2 lta) ltb) // - sg. 
have <-: B = osumf Z i.
  apply: (osumf_exten iN) => j lej; rewrite /Y/Z pa //.
  exact: (clt_ltT lej ha).
have oA: ordinalp A.
   apply: (OS_osumf m1N) => j jl1; apply: aY'.
   rewrite (csumC _ m1); exact:(csum_Mlteq i2N jl1).
have oB: ordinalp B.
   apply: (OS_osumf iN) => j jl1; apply: aY; exact (clt_ltT jl1 ltin).
have fis: Vf f i <c n by apply/(NltP nN); Wtac.
have fil2s: Vf f il2 <c n by apply/(NltP nN); Wtac.
have fi1s: Vf f i1 <c n by apply/(NltP nN); Wtac.
have da:= (hv _ p1 fis).
have dc:= (hv _ p3 fil2s); rewrite -/c in dc.
move: (hu _(p2 _ (cle0n lN))).
rewrite (csum0r (CS_nat iN)); rewrite -/b => db.
have d1a: odegree b <o odegree a by ue.
have d1c: odegree b <o odegree c by ue.
move: (hw _ fis) (hw _ fil2s)(hw _ fi1s); rewrite -/a -/b -/c => ap cp bp.
move: (proj32_1 ap)(proj32_1 bp)(proj32_1 cp) => oa ob oc.
have hc: l +c (csucc (csucc i)) = il2.
  by rewrite csumC (csum_Sn _ i1N) (csum_Sn _ iN).
have [eq1 eq2]: (A +o a = a /\ A +o c = c).
  have: m1 <=c l.
   apply:(csum_le2l i2N m1N lN).
   rewrite m1v (csumC _ l) hc; apply:cleR; fprops.
  rewrite /A; move:m1N; move:(m1); apply: Nat_induction.
    by move => _; rewrite osum_f0 (osum0l oa) (osum0l oc).
  move => q qN Hrec sql.
  have ql:= (cleT (cleS qN) sql).
  move:(Hrec ql) => [eqa eqb].
  set d := (Y (q +c csucc (csucc i))).
  have oD: ordinalp (osumf (fun z => Y (z +c csucc (csucc i))) q).
    apply: (OS_osumf qN) => j jlt; apply:aY.
    have jlt1:=(clt_leT jlt ql).
    move:(csum_Mlteq i2N  jlt1); rewrite hc => hd;apply:(clt_ltT hd ltil2n).
  have ql': q <c l by apply/(cleSltP qN).
  have lt2: q +c csucc (csucc i) <c n.
     move:(csum_Mlteq i2N ql'); rewrite hc => hd; exact(clt_ltT hd ltil2n).
  have lt3: Vf f (q +c csucc (csucc i)) <c k.
    have eq3: q +c (csucc (csucc i)) = csucc (i +c (csucc q)).
      rewrite (csum_nS _ i1N) csumC. 
      by rewrite (csum_Sn _ iN) (csum_nS _ qN).
    by move: (p2 _ sql); rewrite - eq3. 
  move: (hu _ lt3); rewrite -/(Y _ ) -/d => dd.
 have dp: \0o <o d.
    by apply: hw; apply /(NltP nN); Wtac; rewrite sf;apply /(NltP nN).
  have ot:= proj32_1 dp.
  have d2a: odegree d <o odegree a by ue.
  have d2c: odegree d <o odegree c by ue.
  rewrite !(osum_fS _ qN) - (osumA ot oD oa) eqa.
  rewrite - (osumA ot oD oc) eqb (ord_negl_p7 dp ap d2a).
  by rewrite (ord_negl_p7 dp cp d2c).
rewrite (osumA ob oa oB) (ord_negl_p7 bp ap d1a)  (osumA oA oa oB) eq1.
rewrite (osumA oc oa oB) (osumA oA (OS_sum2 oc oa) oB) (osumA oA oc oa) eq2.
rewrite (osumA ob (OS_sum2 oc oa) oB) (osumA ob oc oa) (ord_negl_p7 bp cp d1c).
done.
Qed.

Lemma nds_fL_prop4 f:
  inc f (permutations n) -> exists g,
  [/\ inc g (permutations n), nds_sc n X f = nds_sc n X g,
      fL f  = fL g &
      forall i, i<c n -c k -> pL g ((fL g) +c i) ].
Proof.
move => fp.
move: (NS_diff k nN) (cleR (CS_diff n k)).
move:{1 2 4} (n -c k); apply: Nat_induction. 
  by move => _; exists f; split => // i /clt0.
move => q qN Hrec le1.
move:(cleS qN) => le2.
move: (Hrec (cleT le2 le1)) => [g [ga gb gc gd]].
move: (nds_fL_prop ga) => [aa ab ac ad ae].
have kN:= (NS_lt_nat kln nN).
move: (csum_Meqlt kN  (clt_leT (cltS qN) le1)).
rewrite (cdiff_pr (proj1 kln))=> lt2.
move: (cle_ltT (csum_Mleeq q ab) lt2) => lt3.
case:(nds_fL_prop2 ga qN lt3 gd).
+ by move => [ua _]; move: le1 (cltS qN); rewrite - ua => sa /(cleNgt sa).
+ move => h; exists g; split; [ exact | exact| exact | ].
  move => j; move/(cltSleP qN) => ljq.
  case: (equal_or_not j q) => ejq; [ ue | by apply: gd].
set q1:= fL g +c q.
move => [q1p [ j [jN ja jb jc jd]]].
move: (cpred_pr (NS_sum aa qN) q1p); rewrite -/q1; move=> [je ef].
have eg:=(csum_Sn j je).
have sa: csucc (csucc (cpred q1 +c j)) <c n by rewrite -eg - ef.
have sc: pL g (csucc (csucc (cpred q1 +c j)))  by rewrite -eg - ef.
have sb: (forall j0 : Set, j0 <=c j -> pS g (csucc (cpred q1 +c j0))). 
  by move => j1 jl1; rewrite -(csum_Sn _ je) - ef; apply: jd.
move:(nds_fL_prop3 ga je jN sa jc sb sc) => [g1 [r1 r2 r3 r4]].
rewrite -ef in r3 r4.
have sw: fL g = fL g1.
  move: (nds_fL_prop r1) => [aa' _ _ ad' ae'].
  have af: fL g <=c q1 by apply /csum_M0le; fprops.
  have q1N: natp q1 by move:(NS_succ je); rewrite - ef.
  case: (NleT_el aa' q1N) => ag; last first. 
    by move:(ae' _ ag) r4; rewrite /pS/pL => l1 /(cltNge l1).
  case:(NleT_ell aa aa') => l3.  
  + exact.
  + move: (ae' _ l3) ad; rewrite /pL/pS (r3 _ (clt_leT l3 ag)).
    move => l1 l2; case: (cltNge l1 l2).
  + move: (ae _ l3) ad'; rewrite /pL/pS (r3 _ (clt_leT l3 af)).
    move => l1 l2;case: (cltNge l1 l2).
exists g1;split; [exact | ue| by rewrite gc | ].
move => i /(cltSleP qN) isa; rewrite - sw.
case:(equal_or_not i q) => eqi; first by rewrite eqi.
move: (csum_Meqlt aa (conj isa eqi));rewrite /pL => /r3 <-. 
apply: (gd _ (conj isa eqi)).
Qed.

Lemma nds_fL_prop5 f:
  inc f (permutations n) -> exists K s1 s2,
  [/\ sub K k, inc s1 (permutations (cardinal K)), 
     inc s2 (permutations (n -c k)) &
     nds_sc n X f = nds_sc (n-c k) (fun z => X (k +c z)) s2 +o
       nds_sc_Ktau X K s1].
Proof.
move/nds_fL_prop4 => [g [gp -> _ p1]]. rewrite /nds_sc_Ktau.
move:(nds_fL_prop gp) => [qN p2 p3 p4 p5].
move/permutationsP:gp =>[[injg [fg sjg]] sg tg].
rewrite sg tg in sjg.
set q := (fL g); rewrite -/q in qN p1 p2 p3 p4 p5.
have nkN:=(NS_diff k nN).
have le3 := (cdiff_pr3 nN p2 (proj1 kln)).
have eq2:=(cdiff_pr (proj1 p3)). 
have eq3:=(cdiff_pr le3).
have eq4:=(cdiff_pr (proj1 kln)).
have nqN:= (NS_diff q nN).
have qnkN:= (NS_sum qN nkN).
have kN:= (NS_lt_nat kln nN).
have lt1: q +c (n -c k) <=c n.
  move: (csum_Mleeq n p2); rewrite - {1} eq4 (csumC k) csumA (csumC _ k).
  by move/(csum_le2l kN qnkN nN).
have sK: sub q (source g) by rewrite sg; exact (proj33 (proj1 p3)).
move: (cardinal_image sK injg);set K := Vfs g q.
rewrite (card_nat qN) => cK.
have iKP: forall i, inc i K <-> exists2 j, inc j q & i = Vf g j.
  move => i; split; by move /(Vf_image_P fg sK). 
have h0:sub K k.
 by move => t /iKP [j jE ->]; apply/(NltP kN); apply: p5; apply/(NltP qN).
set E1:= (q +c (n -c k)) -s q.
set K1 := Vfs g E1.
have sK1: sub E1 (source g).
  move => t /setC_P [/(NltP qnkN) ha _]; rewrite sg; apply/(NltP nN).
  exact: (clt_leT ha lt1).
have iK1P: forall i, inc i K1 <-> exists2 j, inc j E1 & i = Vf g j.
  move => i; split; by move /(Vf_image_P fg sK1). 
have cK1: cardinal K1 = n -c k.
  move: (cardinal_image sK1 injg); rewrite -/K1.
  have ha:=(proj33 (Nsum_M0le (n-c k) qN)).
  rewrite  (cardinal_setC4 ha (finite_set_nat qnkN)).
  by rewrite (card_nat qnkN) (card_nat qN) csumC (cdiff_pr1 nkN qN).
have sK1': sub K1  (n -s k).
  move => t/iK1P [j ja ->]; apply/setC_P; split.
    by move: (Vf_target fg (sK1 _ ja)); rewrite tg.
  have jN: natp j by move:(sK1 j ja); rewrite sg => /(NS_inc_nat nN).  
  move/setC_P:ja => [/(NltP qnkN) ja jb]; dneg h.
  case: (cleT_el (CS_nat qN) (proj31_1 ja)); last by move/(NltP qN).
  move => le1; move:(cdiff_pr le1) => eqa.
  rewrite - eqa in ja.
  move:(csum_lt2l qN (NS_diff q jN) nkN ja) => jc.
  move: (p1 _ jc); rewrite eqa /pL => jd.
  by move/(NltP kN):h => /(cleNgt jd).
case: (equal_or_not  K1 (n -s k))=> eq5; last first.
  move: (proj33 (proj1 kln)) (finite_set_nat nN) => s1 fn.
  move: (sub_finite_set (@sub_setC n k) fn) => ha.
  move:(proj2 (strict_sub_smaller (conj sK1' eq5) ha)).
  by rewrite (cardinal_setC4 s1 fn) (card_nat kN) (card_nat nN) cK1.
have p6: forall i, i<c n -> q +c (n -c k) <=c i -> pS g i.
   move => i li1 li2; rewrite /pS; ex_middle h.
   have isn: inc i (source g) by rewrite sg; apply/(NltP nN).
   have: inc (Vf g i) K1.
      rewrite eq5; apply/setC_P; split; [Wtac | by move/(NltP kN)].
  move/(iK1P) => [j /setC_P [/(NltP qnkN) ja jb] eqa].
  have jsg: inc j (source g).
    by move: (clt_ltT (clt_leT ja li2) li1) => /(NltP nN); rewrite sg.
  move: ((proj2 injg) i j isn jsg eqa) => eij; rewrite - eij in ja.
  case: (cltNge ja li2).
have sKb: sub K n by move => t /iKP [ u /sK usf ->]; Wtac.
have sKN:=(sub_trans sKb (NS_inc_nat nN)).
have fsK:=(sub_nat_finite nN sKb).
move:(nth_elt_bf sKN fsK) ; set FK:= Lf _ _ _; move => [pa pb].
have sFK: source FK = q by rewrite /FK cK; aw.
have tFK: target FK = K by rewrite /FK cK; aw.
have ax2: lf_axiom (nth_elt K) q K.
  move => t /(NltP qN) lt2; move:(lt2); rewrite - cK => lt3.
  by move:(nth_set7 (NS_lt_nat lt2 qN) sKN lt3) => [/setC_P []].
pose s1' i := (Vf (inverse_fun FK) (Vf g i)).
have ax1: lf_axiom s1' q q.
  move => t ta;rewrite - sFK; apply:(inverse_Vis pa); rewrite tFK.
  by apply /iKP; exists t. 
have pc: forall i, inc i q -> Vf FK (s1' i) = Vf g i. 
  move => i iE; rewrite /s1' (inverse_V pa) // tFK; apply/iKP; ex_tac.
set s1 := Lf s1' q q .
have s1fb: bijection s1.
   apply:lf_bijective; first by exact.
      move => u v uE vE eq6.
      move: (f_equal (Vf FK) eq6); rewrite (pc _ uE) (pc _ vE) => eq7.
      exact: (proj2 injg _ _ (sK _ uE) (sK _ vE) eq7).
   rewrite - sFK;move => y ye.
   move: (Vf_target (proj1(proj1 pa)) ye); rewrite tFK => /iKP [u us eq1].
   by rewrite - sFK in us; ex_tac; rewrite /s1'- eq1 (inverse_V2 pa ye).
have s1perm: inc s1 (permutations q).
  apply/permutationsP; rewrite /s1; split; aw. 
set Ec:= (n -c k).
pose s2' z := (Vf g (q +c z) -c k).
have pd1: forall z, z <c n -c k -> inc (q +c z) (source g).
  move => z zi.
  move:(csum_Meqlt qN (clt_leT zi le3)).
  by rewrite eq2 => /(NltP nN);rewrite - sg => ha.
have ax3: lf_axiom s2' Ec Ec.
  move => z /(NltP nkN)  zi.
  move: (Vf_target fg (pd1 _ zi)); rewrite tg  => /(NltP nN) lt2.
  rewrite -/q - cK in lt2.
  move:(p1 _ zi);rewrite /pL /s2' - cK=> fzi.
  apply/(NltP nkN); exact(cdiff_pr7 fzi lt2 nN).
have pd: forall z, z <c n -c k -> (s2' z) +c k = Vf g (q +c z).
  move => z /p1;rewrite /pL /s2' - cK=> fzi.
  by rewrite csumC; apply:cdiff_pr. 
set s2 := Lf s2' Ec Ec.
have s2fb: bijection s2.
  apply:bijective_if_same_finite_c_inj.
  + rewrite /s2; aw.
  + by rewrite /s2/Ec; aw; apply:finite_set_nat.
  + apply: lf_injective => // u v /(NltP nkN) ue /(NltP nkN) ve eq.
    move: (f_equal (fun z => z +c k) eq).
    rewrite (pd _ ue) (pd _ ve) => eq7.
    move:(proj2 injg _ _ (pd1 _ ue)(pd1 _ ve) eq7).
    exact:(csum_eq2l qN (NS_lt_nat ue nkN) (NS_lt_nat ve nkN)).
have s2perm: inc s2 (permutations (n -c k)).
  apply/permutationsP; rewrite /s2; split; aw.
exists K, s1, s2; rewrite cK;split => //. 
rewrite /nds_sc_Ktau.
set X1:=osumf (fun z => X (nth_elt K (Vf s1 z))) q.
have r1: X1 = osumf (fun z => X (Vf g z)) q.
  apply: (osumf_exten qN) => i /(NltP qN) iE.
  have vK: inc (Vf g i) (target FK) by rewrite tFK;apply/iKP; ex_tac.
  rewrite /s1/s1'; aw; rewrite  -{2} (inverse_V pa vK) {2} /FK cK.
  aw; rewrite - sFK; apply: (inverse_Vis pa vK).
set X2:= osumf (fun z=> X (Vf g (z +c q))) (n -c k).
have r2: X2 = osumf (fun z => X (k +c Vf s2 z)) (n -c k).
   apply:(osumf_exten nkN).
   move => z znk /=;rewrite (csumC z) (csumC k) - (pd _ znk) /s2; aw.
   by apply/(NltP nkN).
move: Xax => [_ oe hw hu hv].
have aX: nds_ax (fun z => X (Vf g z)) n. 
   move => z /(NltP nN) zi. 
   rewrite - sg in zi.
   by move:(Vf_target fg zi); rewrite tg => /(NltP nN) /hw /proj32_1.
have aX2: ord_below_n (fun z=> X (Vf g (z +c q))) (n -c q).
  move=> z zle1; apply: aX; move: (csum_Mlteq qN zle1).
    by rewrite (csumC (_ -c _)) eq2.
rewrite /nds_sc - {1} eq2;rewrite - eq2 in aX.
rewrite - eq3 in aX2.
rewrite (osum_fA qN nqN aX) -/X1 - r1 - eq3.
rewrite (osum_fA nkN (NS_diff (n -c k) nqN) aX2) - r2.  congr (_ +o X1).
have knz1:= (cdiff_nz kln).
move: (cpred_pr nkN knz1) => [nk1N nk1v].
rewrite /X2 nk1v (osum_fS _ nk1N).
set X3:= osumf _ _.
set X4:= osumf _ _.
set a := X _.
move:(cltS nk1N); rewrite -nk1v => /p1; rewrite /pL csumC => le1.
have lt2: inc ((cpred (n -c k) +c q)) (source g).
  rewrite csumC;apply/pd1; rewrite {2} nk1v; apply:(cltS nk1N).
have lt3: Vf g (cpred (n -c k) +c q) <c n.
  by apply/(NltP nN); rewrite - {2}tg; Wtac.
move: (hw _ lt3) (hv _ le1 lt3); rewrite -/a => ap da.
have oa:= proj32_1 ap.
have oX4: ordinalp X4. 
  apply:(OS_osumf nk1N) => t ta;apply: aX2.
  move: (clt_leT ta (cleS nk1N)); rewrite - nk1v => sa.
  exact: (clt_leT sa (Nsum_M0le ((n -c q) -c (n -c k)) nkN)).
set wa:=((n -c q) -c csucc (cpred (n -c k))).
have wN: natp wa by rewrite /wa; fprops.
have wa1: (wa +c (n -c k)) =  (n -c q).
  by rewrite /wa -nk1v (csumC (_ -c _)) cdiff_pr. 
have oX3: ordinalp X3. 
  apply:(OS_osumf wN) => z zw;apply: aX2;rewrite eq3.
  by move: (csum_Mlteq nkN zw); rewrite  -nk1v wa1.
rewrite (osumA oX3 oa oX4); congr (_ +o X4).
rewrite /X3 -/wa - nk1v.
have eq7: ((wa +c (n -c k)) +c q) = n by rewrite  wa1 csumC eq2//.
have: wa +c ((n -c k) +c q) <=c n by rewrite csumA eq7; fprops.
clear eq7 wa1.
move: wa wN; apply: Nat_induction.
  by move => _ ; rewrite osum_f0 (osum0l oa).
move=> j jN Hrec.
have sN:= NS_sum jN (NS_sum nkN qN).
rewrite (csum_Sn _ jN) => /(cleSltP sN) ha.
have oS: ordinalp (osumf (fun z => X (Vf g ((z +c (n -c k))
      +c q))) j).
  apply:(OS_osumf jN) => z zl; apply:aX2;rewrite eq3.
  move:(csum_Mlteq nkN zl) => hc.
  move:(Nsum_M0le (j +c (n -c k)) qN); rewrite csumC => hd.
  rewrite csumA in ha; move: (cdiff_pr7 hd ha nN).    
  rewrite (cdiff_pr1 (NS_sum jN nkN) qN) => he; exact: (clt_ltT hc he).
rewrite (osum_fS _ jN) - csumA.
set j0 := j +c ((n -c k) +c q).
have j0b: inc j0 (source g) by rewrite sg; apply/(NltP nN).
rewrite -/j0 in ha.
have hb:q +c (n -c k) <=c j0. 
  rewrite /j0 csumC (csumC j); apply:csum_M0le; fprops.
move:(Vf_target fg j0b); rewrite tg => /(NltP nN) => hc.
have ofp:= hw _ hc.
have oft:= proj32_1 ofp.
rewrite - (osumA oft oS oa) (Hrec (proj1 ha)).
move: (hu _ (p6 _ ha hb)) => hd.
rewrite - hd in da; exact: (ord_negl_p7 ofp ap da).
Qed.


Lemma nds_fL_prop6 v (X2:= (fun z => X (k +c z))):
  inc v (nds_sums n X) -> 
  exists v1 v2, [/\ inc v1 (nds_tn_S X k), inc v2 (nds_sums (n -c k) X2) &
   v = v2 +o v1].
Proof. 
move => /funI_P [z zp ->].
move: (nds_fL_prop5 zp) => [K [s1 [s2 [/setP_P pa pb pc ->]]]].
set v1:=nds_sc (cardinal K) (fun z0 => X (nth_elt K z0)) s1.
set v2:=nds_sc (n -c k) (fun z => X (k +c z)) s2.
have v2p:inc v2 (nds_sums (n -c k) X2) by apply/funI_P; ex_tac.
exists v1, v2; split => //;apply/setUf_P; ex_tac; apply/funI_P; ex_tac.
Qed.


Lemma nds_tg9 (X2:= (fun z => X (k +c z))):
  (nds_sums n X) = 
     unionf (nds_tn_S X k) (fun v1 => fun_image (nds_sums (n -c k) X2) 
         (fun v2 => v2 +o v1)).
Proof.
set_extens v.
  move /nds_fL_prop6 => [v1 [v2 [va vb vc]]].
  apply/setUf_P; ex_tac; apply/funI_P; ex_tac.
move/setUf_P => [v1 /setUf_P [K /setP_P Kp/funI_P [s1 s1p va]]].
move => /funI_P [v2 /funI_P [s2 s2p ->]] => ->; rewrite va; apply/funI_P.
set K' := k -s K.
set q := cardinal K.
have kN:=(NS_lt_nat kln nN).
move:(sub_smaller Kp); rewrite (card_nat kN); rewrite -/q => leqk.
have lqn:=(cle_ltT leqk kln).
have qN:= NS_lt_nat lqn nN.
have nkN:=(NS_diff k nN).
have qnN:= (NS_sum qN nkN).
have sKN:= (sub_trans Kp (NS_inc_nat kN)).
have sK'N: sub K' Nat by  exact: (sub_trans (@sub_setC k K) (NS_inc_nat kN)).
have cK': cardinal K' = k -c q.
  by rewrite (cardinal_setC4 Kp (finite_set_nat kN)) (card_nat kN).
have qqn:=(Nsum_M0le (n -c k) qN).
have kqN:= (NS_diff q kN).
have q'N: natp (cardinal K') by rewrite cK'.
have knk:=(cdiff_pr (proj1 kln)).
set qn := q +c (n -c k); rewrite -/qn in qnN qqn.
have qnkq: qn +c (k-c q) = n.
  by rewrite -knk /qn (csumC q) - csumA (cdiff_pr leqk) csumC.
have lqnn: qn <=c n by rewrite - qnkq; apply:(Nsum_M0le _  qnN).
have sBkBn:=(proj33 (proj1 kln)).
move/permutationsP: (s1p) => [bs1 ss1 ts1]; move: (proj1(proj1 bs1)) => fs1.
move/permutationsP: (s2p) => [bs2 ss2 ts2]; move: (proj1(proj1 bs2)) => fs2.
pose s i := Yo (i <c q) (nth_elt K (Vf s1 i)) 
     (Yo (i <c qn) (Vf s2 (i -c q) +c k) (nth_elt K' (i -c qn))). 
have pa: forall i, i<c q -> s i = (nth_elt K (Vf s1 i)).
   by move => i liq; rewrite /s; Ytac0.
have pb: forall i, i<c q -> inc (s i) K.
   move => i iq; rewrite (pa _ iq).
   have p1: inc (Vf s1 i) q by rewrite /q; Wtac; rewrite ss1; apply/(NltP qN).
   move/(NltP qN): p1 => ha. 
   apply:(nth_elt_inK (NS_lt_nat ha qN) sKN ha).
have pc: forall i, qn <=c i -> s i =  (nth_elt K' (i -c qn)).
   rewrite /s => i le1;Ytac h; first by case:(cleNgt qqn (cle_ltT le1 h)).
   by rewrite (Y_false (cleNgt le1)).
have pd: forall i, qn <=c i -> i <c n -> inc (s i) K'.
  move => i l1 l2; rewrite (pc _ l1); move:(NS_lt_nat l2 nN) => iN.
  apply:(nth_elt_inK (NS_diff qn iN) sK'N); rewrite cK'. 
  by apply:(cdiff_Mlt kqN iN l1); rewrite csumC qnkq.
have pe: forall i, q <=c i -> i <c qn -> s i = (Vf s2 (i -c q) +c k).
  by move => i l1 l2; rewrite /s (Y_true l2) (Y_false (cleNgt l1)). 
have pf: forall i, q <=c i -> i <c qn ->
  [/\ s i = (Vf s2 (i -c q) +c k), inc (Vf s2 (i -c q)) ((n -c k)) &
   inc (s i) (n -s k)].
  move => i l1 l2; rewrite (pe i l1 l2).
  have iN:= NS_lt_nat l2 qnN. 
  rewrite - {1}(cdiff_pr l1) /qn in l2.
  have sa:= (csum_lt2l qN (NS_diff q iN) nkN l2).
  have sb: inc (Vf s2 (i -c q)) (n -c k). 
    by Wtac; rewrite ss2; apply/NltP.
  move/(NltP nkN):(sb) => sc.
  move:(csum_Mlteq kN sc); rewrite (csumC (n -c k)) knk => sd.
  split => //; apply/setC_P; split; first by apply /(NltP nN).
  move/(NltP kN); move:(Nsum_Mle0 (Vf s2 (i -c q)) kN). 
  move =>  ua ub; case: (cltNge ub ua).
have ax: forall i, inc i n -> inc (s i) n.
  move => i /(NltP nN) lin.
  case: (cleT_el  (CS_nat qN) (proj31_1 lin)) => le1.
    case: (cleT_el (CS_nat qnN) (proj31_1 lin)) => le2.
      by move: (pd i le2 lin) => /setC_P [/sBkBn ha _].
    by move:(pf _ le1 le2) =>[_ _ /setC_P []].
  by move: (pb i le1) =>  /Kp /sBkBn.
have sjs: forall j, inc j n -> exists2 i, inc i n &  j = s i. 
  move => j /(NltP nN) ljn.
  have jN:= (NS_lt_nat ljn nN).
  case:(cleT_el (CS_nat kN) (proj31_1 ljn)) => lkj.
    have l1: (j -c k)  <c (n-c k).
      rewrite - (cdiff_pr lkj) - knk in ljn.
      exact: (csum_lt2l kN (NS_diff k jN) nkN ljn).
    move: (perm_int_surj nkN s2p l1) => [i1 l2 iv].
    have i1N := NS_lt_nat l2 nkN.
    have l3:=(csum_Meqlt qN l2).
    move:(proj31 (pf _ (Nsum_M0le i1 qN) l3)).
    rewrite {2} csumC (cdiff_pr1 i1N qN) iv (csumC _ k) (cdiff_pr lkj) => <-.
    exists (q +c i1) => //; apply/(NltP nN); exact:(clt_leT l3 lqnn).
  case: (inc_or_not j K) => jK.
      move:(nth_elt_surj sKN qN jK) => [i ia ->].
      move:(perm_int_surj qN s1p ia) => [i1 ib <-].
      rewrite - (pa _ ib); exists i1 => //; apply/(NltP nN). 
      exact:(clt_ltT ib lqn).
  have jK': inc j K' by apply/setC_P; split => //; apply/(NltP kN).
  move:(nth_elt_surj sK'N q'N jK') => [i ia ->].
  rewrite cK' in ia.
  move: (csum_Meqlt qnN ia); rewrite qnkq => /(NltP nN) ic; ex_tac.
  rewrite (pc _ (Nsum_M0le i qnN)) csumC cdiff_pr1 //.
  apply:(NS_lt_nat ia kqN).
set sigma := Lf s n n.
have sp: inc sigma (permutations n).
   rewrite /sigma; apply/permutationsP; hnf; aw; split => //.
   apply:bijective_if_same_finite_c_surj; aw; first by apply:finite_set_nat.
   by apply: lf_surjective.
ex_tac.
move: Xax => [qa qb qc qd qe].
have: ord_below_n (fun z => X (Vf sigma z)) n.
  rewrite /sigma;move => y /(NltP nN) ys; aw.
  by move:(ax _ ys) => /(NltP nN) /qc [[]].
rewrite /nds_sc - {1 3}qnkq => ax2.
have ax3: ord_below_n (fun z=> X (Vf sigma z)) (q +c (n -c k)).
 move => z z1; apply: ax2; rewrite qnkq;  exact:(clt_leT z1 lqnn).
rewrite (osum_fA qnN kqN ax2)  (osum_fA qN nkN ax3) /nds_sc_Ktau/nds_sc.
set A := osumf _ _; set B := osumf _ _; set C' := osumf _ _.
set A' := osumf _ _; set B' := osumf _ _.
have ea: A = A'. 
  apply: (osumf_exten nkN) => i l1.
  have iB:= NS_lt_nat l1 nkN.
  move: (csum_Mlteq qN l1) ; rewrite (csumC (n -c k))- /qn => l2.
  have ha: inc (i +c q) n by apply /(NltP nN); exact:(clt_leT l2 lqnn).
  move:(Nsum_M0le i qN); rewrite csumC => hb.
  by rewrite /sigma; aw; rewrite (pe _ hb l2) /X2 csumC cdiff_pr1.
have eb: B = B'. 
  apply: (osumf_exten qN) => i l1.
  have ib:inc i n by apply /(NltP nN); exact:(clt_ltT l1 lqn).
   by rewrite /sigma; aw; rewrite - ( pa _ l1).
have oa: ordinalp A'.
   apply:(OS_osumf nkN) => i l1; apply: ax3.
   by move: (csum_Mlteq qN l1);rewrite (csumC q).
have ob: ordinalp B'.
   apply:(OS_osumf qN) => i l1; apply: ax2; rewrite qnkq.
   exact:(clt_ltT l1 lqn).
have oc: ordinalp C'.
   apply:(OS_osumf kqN) => i l1; apply: ax2. 
   by rewrite (csumC qn);apply:csum_Mlteq. 
rewrite (osumA oc oa ob) - ea; apply: (f_equal2 _ _ eb).
have knz1:= (cdiff_nz kln).
move: (cpred_pr nkN knz1) => [nk1N nk1v].
rewrite /A nk1v (osum_fS _ nk1N) /X2.
set C:= osumf _ _ .
move: (cltS nk1N); rewrite - nk1v => eq1.
have oc': ordinalp C.
  apply:(OS_osumf nk1N) => i lin.
  have ha:inc i (source s2). 
   rewrite ss2; apply/(NltP nkN); exact (clt_ltT lin eq1).
  move:(Vf_target fs2 ha); rewrite ts2 => /(NltP nkN) lt1.
  by move:(csum_Meqlt kN lt1); rewrite knk => /qc /proj32_1.
have ha:inc (cpred (n -c k)) (source s2).
   by rewrite ss2; apply/(NltP nkN); rewrite {2} nk1v; apply: cltS.
move:(Vf_target fs2 ha); rewrite ts2 => /(NltP nkN) lt1.
move:(csum_Meqlt kN lt1); rewrite knk => hb.
move:(Nsum_M0le (Vf s2 (cpred (n -c k))) kN) => hc.
move: (qc _ hb) (qe _ hc hb);set a:= X _ => ap da.
rewrite (osumA oc (proj32_1 ap) oc'); congr (_ +o C).
rewrite /C'.
have hd: forall i, i<c k -c q ->  (Vf sigma (i +c qn)) <c k.
  move => i lik.
  move:(csum_Meqlt qnN lik); rewrite qnkq => l1.
  move: (Nsum_M0le i  qnN) => l2.
  move/(NltP nN):(l1) => iN.
  by rewrite csumC /sigma; aw; move: (pd _ l2 l1) => /setC_P []/(NltP kN).
have he:  forall i, i<c k -c q -> let b := X (Vf sigma (i +c qn)) in
  [/\ ordinalp b, \0o <o b & odegree b = e].
  move => i l1; move:(hd _ l1) => sa.
  move:(qd _ sa) (qc _ (clt_ltT sa kln)) => r1 r2.
  split; [exact (proj32_1 r2) | exact r2| exact r1].
symmetry.
move:kqN (cleR (CS_nat kqN)).
have oa':=(proj32_1 ap).
move: {-3} (k -c q); apply: Nat_induction.
  by rewrite osum_f0 (osum0l oa').
move => i iN Hrec le1.
rewrite (osum_fS _ iN).
move/(cleSltP iN): (le1) => /he [].
set b :=  (X (Vf sigma (i +c qn)))  => ob' bp db.
have ikq:=(cleT (cleS iN) le1).
have oc'':ordinalp (osumf (fun z : Set => X (Vf sigma (z +c qn))) i).
   apply: (OS_osumf iN) => j jli.
   by move: (he _ (clt_leT jli ikq)) => [].
rewrite - db in da.
rewrite - (osumA ob' oc'' oa').
by rewrite (Hrec ikq) (ord_negl_p7  bp ap da).
Qed.

Lemma nds_tg10 : nds_card n X <=c (ndsC k) *c (nds_F (n -c k)).
Proof.
rewrite /nds_card nds_tg9. 
set F:= (fun v1 : Set => _).
set E :=  (nds_tn_S X k).
move:(csum_pr1_bis E F).
set g1:= (Lg E (fun a => cardinal (Vg (Lg E F) a))).
have <-: csum g1 = csumb E  (fun a => cardinal (F a)).  
  apply:csumb_exten => t tx; bw.
set g2 := cst_graph E (nds_F (n -c k)).
have sd: domain g1 = domain g2 by rewrite /g1/g2; bw.
have kN:= NS_lt_nat kln nN.
have nkN:= (NS_diff k nN).
move: (nds_f_def nkN) => [pe _ pf pg].
move: Xax => [pa pb pc pd _].
suff le1:forall x, inc x (domain g1) -> Vg g1 x <=c Vg g2 x.
  move:(csum_increasing sd le1).
  rewrite csum_of_same cprodC - cprod2cl /E -/(nds_tn_C _ _) => ha hb.
  have ax:  nds_tn_ax X k.  
    by split; [move => i ik; apply:pc (clt_ltT ik kln) | exists e].
  move: (cprod_Mlele (nds_tn_max kN knz ax) (cleR (CS_nat pe))) => hc.
  exact:(cleT (cleT hb ha) hc).
rewrite /g1 /g2; bw => a abn; bw. 
have /pg:nds_ax(fun z => X (k +c z))  (n -c k).
  move => i ink; move:(csum_Meqlt kN ink).
  by rewrite(cdiff_pr (proj1 kln)) => /pc /proj32_1.
apply:cleT; apply:fun_image_smaller.
Qed.



Lemma nds_tg11 Y : 
  (forall i, i <c n -c k ->
      \0o <o Y i /\ X (k +c i) = omega0 ^o (osucc e) *o (Y i)) ->
  nds_tn_C X k = ndsC k ->
  nds_card (n-c k) Y = nds_F (n -c k)  ->
  nds_card n X = (ndsC k) *c (nds_F (n -c k)).
Proof.
move => yp <- <-.
rewrite /nds_card nds_tg9. 
have kN := NS_lt_nat kln nN.
have nkN := NS_diff k nN.
move: Xax => [Xpa Xpb Xpc Xpd Xpe].
set c:= omega0 ^o osucc e.
have cp: \0o <o c by apply: (oopow_pos  (OS_succ Xpb)).
have oc:= proj32_1 cp.
set E := (nds_sums (n -c k) (fun z => X (k +c z))).
have ax0: nds_ax Y (n -c k) by move => i /yp [] /proj32_1 h _.
have aux: forall i, inc i (nds_tn_S X k) -> i <o c.
  move => i /setUf_P [K /setP_P K1 /funI_P [s sp] ->].
  have sKN:= (sub_trans K1 (NS_inc_nat kN)).
  have fsk:= (sub_finite_set K1 (finite_set_nat kN)).
  have cKN: natp (cardinal K) by apply /NatP.
  have : forall i, i <c (cardinal K) -> X (nth_elt K (Vf s i)) <o c.
    move => j jk.
    have r0: (Vf s j) <c cardinal K.
      move /permutationsP: sp => [[[fs _] _] ss ts].
      by apply/(NltP cKN); Wtac; rewrite ss; apply/(NltP cKN).
    have r1:(nth_elt K (Vf s j)) <c k.
      apply /(NltP kN); apply:K1; apply: nth_elt_inK => //.
      apply: (NS_lt_nat r0 cKN).
    by move: (proj2(the_CNF_p4 (Xpc _ (clt_ltT r1 kln)))); rewrite Xpd. 
  rewrite /nds_sc_Ktau/nds_sc.
  move: (cardinal K) cKN; apply: Nat_induction; first by rewrite osum_f0.
  move => m mN Hrec aux; rewrite /nds_sc (osum_fS _ mN).
  have oic:= (indecomp_prop4 (OS_succ Xpb)).
  have alc:=(aux _ (cltS mN)).
  apply: (indecomp_prop2 alc _ oic); apply: Hrec => j jm; apply:aux.
  exact: (clt_ltT jm (cltS mN)).
have aux2: forall i j u v, inc i (nds_tn_S X k) -> inc j (nds_tn_S X k) ->
   inc u E -> inc v E ->   u +o i = v +o j -> (u = v /\ i = j).
  move => i j u v ia ja /funI_P  [s1 s1p v1p] /funI_P [s2 s2p v2p].
  move/permutationsP:s1p => [[[fs1 _] _] sf1 tf1].
  move/permutationsP:s2p => [[[fs2 _] _] sf2 tf2].
  have ha: forall q, q<c n -c k -> (Vf s1 q)  <c n -c k.  
    by move => q qa;apply /(NltP nkN); Wtac; rewrite sf1; apply/(NltP nkN).
  have hb: forall q, q<c n -c k -> (Vf s2 q)  <c n -c k.  
    by move => q qa;apply /(NltP nkN); Wtac; rewrite sf2; apply/(NltP nkN).
  have v1a: u =  nds_sc (n -c k) (fun z => c *o Y z) s1.
    by rewrite v1p; apply:(osumf_exten nkN) => q /ha /yp [].
  have v2a: v =  nds_sc (n -c k) (fun z => c *o Y z) s2.
    by rewrite v2p; apply:(osumf_exten nkN) => q /hb/yp [].
  have ax1:nds_ax (fun z => Y (Vf s1 z)) (n -c k).
    by move => z /ha/yp [/proj32_1 hc _].
  have ax2:nds_ax (fun z => Y (Vf s2 z)) (n -c k).
    by move => z /hb/yp [/proj32_1 hc _].
  rewrite  v1a v2a /nds_sc (osumf_mul nkN ax1 oc) (osumf_mul nkN ax2 oc).
  set v1':= osumf _ _; set v2':= osumf _ _ => eq1.
  have ov1: ordinalp v1' by apply:(OS_osumf  nkN); apply: ax1.
  have ov2: ordinalp v2' by apply:(OS_osumf  nkN); apply: ax2.
  move: (aux _ ia) (aux _ ja) => ib jb.
  set a :=  c *o v1' +o i. 
  move:(proj31_1 ib)(proj31_1 jb) => oi oj.
  have oa:ordinalp a by apply (OS_sum2 (OS_prod2 oc ov1) oi).
  by move:(odivision_unique oa oc (And4 ov1 oi (erefl a) ib)
    (And4 ov2 oj eq1 jb)) => [-> rb].
pose F a :=(fun_image E (osum2^~ a)).
have H: forall a, inc a (nds_tn_S X k) -> cardinal (F a) = nds_card (n -c k) Y.
  move => a ias.
  rewrite cardinal_fun_image. 
    rewrite /E -/(nds_card _ _) (nds_card_mul nkN ax0 cp). 
    by apply:(nds_card_exten nkN) => i /yp [_ ->]. 
  move => u v ua va /= eq; exact: (proj1 (aux2 _ _ _ _ ias ias ua va eq)).
rewrite csum_pr4_bis.
  rewrite cprodC /nds_tn_C cprod2cr - csum_of_same /csumb.
  by apply: f_equal; apply: Lg_exten => i ib; apply: H.
move => i j ia ja; case (equal_or_not i j) => eij; [by left | right].
  apply:disjoint_pr => u /funI_P [v1 va vb] /funI_P [v2 vc vd].
  rewrite vb in vd; case eij; exact: (proj2 (aux2 _ _ _ _ ia ja va vc vd)).
Qed.

End NdsStudy.

Lemma nds_s1 n k: natp n -> \0c <c k -> k <c n -> 
   nds_FA n k <=c (ndsC k) *c (nds_F (n -c k)).
Proof.
move => nN kp kln .
move: (cle_ltT (cle0x (proj31_1 kln)) kln) => np.
move: (nds_FA_def nN kp (proj1 kln)).
move=> [pa _ _ [X [pb pc <-]]].  
move:(nds_type_nor_ex nN kp (proj1 kln) pc) => [Y [e [ w ya <-]]].
by apply: (nds_tg10 nN kp kln  ya).
Qed.

Lemma nds_s2 n k: natp n ->  \0c <c k -> k <c n -> 
   nds_FA n k = (ndsC k) *c (nds_F (n -c k)).
Proof.
move => nN kp kln; apply: cleA; first by apply:nds_s1.
move: (cle_ltT (cle0x (proj31_1 kln)) kln) => np.
move: (nds_FA_def nN kp (proj1 kln)) => [pa _ pb _].
have dN:= (NS_diff k nN).
have  [X [ax xa1 xv]]: exists X, [/\ nds_ax X (n -c k),
   forall i, i <c n -c k -> \0o <o X i & nds_card (n -c k) X = nds_F (n -c k)].
   case: (equal_or_not (n -c k) \1c) => ed1.
    exists (fun i => \1o); split => //.
      move => i _;apply: OS1.
      move => i _;apply: olt_01.
    rewrite ed1 (proj2 nds_f01) nds_card_01; fprops.
  have dp: \2c <=c (n -c k) by apply: (cge2 (CS_nat dN) (cdiff_nz kln) ed1).
  move:(proj2(nds_F_FA_def dN dp)) => [k1 [k1a k1b <-]].
  move:(nds_FA_def dN k1a (proj1 k1b)) => [_ _ _ [X [axx [ua _] <-]]].
  by exists X. 
move:(nds_tn_def (NS_lt_nat kln nN)) => [_ _  [Y [ya1 [e oe yd]] yv] _].
set c := omega0 ^o (osucc e).
have cp: \0o <o c by apply:(oopow_pos (OS_succ oe)).
have oc:= (proj32_1 cp).
pose Z i := Yo (i <c k) (Y i) (c *o X (i -c k)). 
have kN:=(NS_lt_nat kln nN).
have nkN:= (NS_diff k nN).
have az1:  forall i, i<c n -> \0o <o  Z i.
  move => i lin; rewrite /Z; Ytac h; first by apply: ya1.
  case: (cleT_el (proj31_1 kln) (proj31_1 lin)) => // leki.
  apply: (oprod2_pos cp (xa1 _ (cdiff_pr7 leki lin nN))).
have az2: forall i, i <c n -> ~ i <c k ->  e <o odegree (Z i).
  move => i lin h.
  rewrite /Z; Ytac0;apply /oltSSP.
  case: (cleT_el (proj31_1 kln) (proj31_1 lin)) => // leki.
  have xip :=(xa1 _ (cdiff_pr7 leki lin nN)).
  have ha:= (oprod_Mle1 oc xip).
  have od:=(OS_degree (proj32 ha)).  
  have:= (ole_ltT ha (proj2 (the_CNF_p4 (oprod2_pos cp xip)))).
  by move /(opow_Meqltr ole_2omega (OS_succ oe) ( (OS_succ od))).
have az3:forall i, i <c k -> odegree (Z i) = e.
  by move => i lik;rewrite /Z; Ytac0;rewrite (yd _ lik). 
have az4:forall i, i <c n -> e <=o odegree (Z i).
  move => i lin; case (p_or_not_p (i <c k)) => h.
    rewrite (az3 _ h); apply: (oleR oe).
   exact: (proj1 (az2 _ lin h)).
have ha: nds_type Z n k.
  split => //; exists e; split => //.
  have ->: (Zo n (fun i => odegree (Z i) = e)) = k.
    set_extens i.
        move => /Zo_P [/(NltP nN) lin di]; apply/(NltP kN).
        by ex_middle h; move:(proj2 (az2 i lin h)); rewrite di; case.
    move => /(NltP kN) => lik; apply:Zo_i.  
      by apply /(NltP nN); move:(clt_ltT lik kln).
    apply: (az3 _ lik).
  by rewrite (card_nat kN).
have hb: nds_type_nor Z n k e.
  split => // i ki lin; apply: (az2 i lin); exact (cleNgt ki).
have hc: forall i,
   i <c n -c k -> \0o <o X i /\ Z (k +c i) = omega0 ^o osucc e *o X i. 
  move => i ib; split; first by apply: xa1.
  rewrite /Z (Y_false (cleNgt (csum_M0le i (proj31_1 kln)))).
  by rewrite csumC (cdiff_pr1 (NS_lt_nat ib nkN) kN).
have hd:nds_tn_C Z k = ndsC k. 
  have sb: same_below Z Y k by move => i  klik; rewrite /Z; Ytac0.
   have ->: nds_tn_C Z k = nds_tn_C Y k by apply:nds_tn_C_exten.
  by rewrite yv (nds_tn_value kN).
have he: nds_ax Z n by move => i lin; exact (proj32_1 (az1 _ lin)). 
by rewrite - (nds_tg11 nN kln hb hc hd xv); apply: pb.
Qed.

Theorem nds_thm_B: 
  nds_fprops nds_F /\
  (forall k, natp k -> card_fifteen <=c k -> 
      nds_F(k +c \5c) = (ndsC \5c) *c nds_F k).
Proof.
have ha: nds_fprops nds_F.
  move: nds_f01  => [p0 p1]; split => //.
  move => n k nN n2 kp lkn.
    rewrite - (nds_s2 nN kp lkn); exact:(proj1 (nds_F_FA_def nN n2) _ kp lkn).
  move => n nN n2.
  move:(proj2 (nds_F_FA_def nN n2)) => [k0 [k0n k0nz <-]].
  by exists k0; rewrite - (nds_s2 nN k0n k0nz).
split; [ exact | by  case: (nds_thm_bourbaki ha)].
Qed.

End NumberSums.

Export NumberSums.
